# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Union

import scipy.stats
import numpy as np

import itertools

from ...core.interfaces import IModel, IDifferentiable, IJointlyDifferentiable
from ...core.acquisition import Acquisition


class ExpectedImprovement(Acquisition):
    def __init__(self, model: Union[IModel, IDifferentiable], jitter: float=0.0)-> None:
        """
        This acquisition computes for a given input the improvement over the current best observed value in
        expectation. For more information see:

        Efficient Global Optimization of Expensive Black-Box Functions
        Jones, Donald R. and Schonlau, Matthias and Welch, William J.
        Journal of Global Optimization

        :param model: model that is used to compute the improvement.
        :param jitter: parameter to encourage extra exploration.
        """

        self.model = model
        self.jitter = jitter

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the Expected Improvement.

        :param x: points where the acquisition is evaluated.
        """

        mean, variance = self.model.predict(x)
        standard_deviation = np.sqrt(variance)
        mean += self.jitter

        y_minimum = np.min(self.model.Y, axis=0)
        u, pdf, cdf = get_standard_normal_pdf_cdf(y_minimum, mean, standard_deviation)
        improvement = standard_deviation * (u * cdf + pdf)

        return improvement

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple:
        """
        Computes the Expected Improvement and its derivative.

        :param x: locations where the evaluation with gradients is done.
        """

        mean, variance = self.model.predict(x)
        standard_deviation = np.sqrt(variance)

        y_minimum = np.min(self.model.Y, axis=0)

        dmean_dx, dvariance_dx = self.model.get_prediction_gradients(x)
        dstandard_deviation_dx = dvariance_dx / (2 * standard_deviation)

        mean += self.jitter
        u, pdf, cdf = get_standard_normal_pdf_cdf(y_minimum, mean, standard_deviation)

        improvement = standard_deviation * (u * cdf + pdf)
        dimprovement_dx = dstandard_deviation_dx * pdf - cdf * dmean_dx

        return improvement, dimprovement_dx

    @property
    def has_gradients(self) -> bool:
        """Returns that this acquisition has gradients"""
        return isinstance(self.model, IDifferentiable)


def get_standard_normal_pdf_cdf(x: np.array, mean: np.array, standard_deviation: np.array) \
        -> Tuple[np.array, np.array, np.array]:
    """
    Returns pdf and cdf of standard normal evaluated at (x - mean)/sigma

    :param x: Non-standardized input
    :param mean: Mean to normalize x with
    :param standard_deviation: Standard deviation to normalize x with
    :return: (normalized version of x, pdf of standard normal, cdf of standard normal)
    """
    u = (x - mean) / standard_deviation
    pdf = scipy.stats.norm.pdf(u)
    cdf = scipy.stats.norm.cdf(u)
    return u, pdf, cdf


class MultipointExpectedImprovement(ExpectedImprovement):
    def __init__(self, model: Union[IModel, IDifferentiable, IJointlyDifferentiable], jitter: float=0.0,
                 fast_compute: bool=False, eps: float=1e-3) -> None:
        """
        This acquisition computes for a given input the improvement over the current best observed value in
        expectation for multiple points. For more information see:

        Efficient Global Optimization of Expensive Black-Box Functions
        Jones, Donald R. and Schonlau, Matthias and Welch, William J.
        Journal of Global Optimization

        Implementation based on papers and their implementations:

        Fast computation of the multipoint Expected Improvement with
        applications in batch selection
        Chevalier, C. and Ginsbourger, D.
        International Conference on Learning and Intelligent Optimization.
        Springer, Berlin, Heidelberg, 2013.

        Gradient of the acquisition is derived in:

        Differentiating the multipoint Expected Improvement for optimal batch design
        Marmin, S., Chevalier, C. and Ginsbourger, D.
        International Workshop on Machine Learning, Optimization and Big Data.
        Springer, Cham, 2015.

        Source code for both: https://github.com/cran/DiceOptim

        :param model: model that is used to compute the improvement.
        :param jitter: parameter to encourage extra exploration.
        :param fast_compute: Whether to use faster approximative method.
        :param eps: Grid length for numerical derivative in approximative method.
        """
        super(MultipointExpectedImprovement, self).__init__(model, jitter)
        self.fast_compute = fast_compute
        self.eps = eps

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Computes the multipoint Expected Improvement.

        :param x: points where the acquisition is evaluated.
        :return: multipoint Expected Improvement at the input.
        """
        mean, variance = self.model.predict_with_full_covariance(x)
        y_minimum = np.min(self.model.Y, axis=0)
        return -self._get_acquisition(mean.flatten(), variance, y_minimum)[0]

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the multipoint Expected Improvement and its derivative.

        :param x: locations where the evaluation with gradients is done.
        :return: multipoint Expected Improvement and its gradient at the input.
        """
        mean, variance = self.model.predict_with_full_covariance(x)
        mean = mean.flatten()
        y_minimum = np.min(self.model.Y, axis=0)
        qei, pk, symmetric_term = self._get_acquisition(mean, variance, y_minimum)

        mean_dx, variance_dx = self.model.get_joint_prediction_gradients(x)
        qei_grad = self._get_acquisition_gradient(mean, variance, mean_dx, variance_dx,
                                                  y_minimum, pk, symmetric_term)
        return -qei, -qei_grad

    def _get_acquisition(self, mu: np.ndarray, Sigma: np.ndarray, y_minimum: float) -> Tuple:
        """
        Computes the multi point Expected Improvement. A helper function for the class.

        :param mu: Prediction mean at locations where the evaluation is done.
        :param Sigma: Prediction covariance at locations where the evaluation is done.
        :param y_minimum: The best value evaluated so far by the black box function.
        :return: Multipoint Expected Improvement where the mean and the covariance are evaluated.
        """
        q = mu.shape[0]
        pk = np.zeros((q,))
        first_term = np.zeros((q,))
        non_symmetric_term = np.zeros((q, q))
        symmetric_term = np.zeros((q, q))
        for k in range(q):
            Sigma_k = get_covariance_given_smallest(Sigma, k)
            mu_k = mu[k] - mu
            mu_k[k] = mu[k]
            b_k = np.zeros((q,))
            b_k[k] = y_minimum

            pk[k] = scipy.stats.multivariate_normal.cdf(b_k - mu_k, np.zeros((q,)), Sigma_k)

            first_term[k] = (y_minimum - mu[k]) * pk[k]
            symmetric_term[k, :], non_symmetric_term[k, :] = self._get_non_symmetric_and_symmetric_term_k(b_k, mu_k, Sigma_k, pk, k)

        # Symmetrify the symmetric term
        symmetric_term = symmetric_term + symmetric_term.T
        symmetric_term[range(q), range(q)] = 0.5 * np.diag(symmetric_term)
        second_term = np.sum(symmetric_term * non_symmetric_term)

        # See equation (3) in the paper for details:
        opt_val = np.sum(first_term) + np.sum(second_term)

        return opt_val, pk, symmetric_term

    def _get_acquisition_gradient(self, mu: np.ndarray, Sigma: np.ndarray, dmu_dx: np.ndarray,
                                  dSigma_dx: np.ndarray, y_minimum: float, pk: np.ndarray,
                                  symmetric_term: np.ndarray) -> np.ndarray:
        """
        Computes the gradient of multi point Expected Improvement. A helper function for the class.

        :param mu: Prediction mean at locations where the evaluation is done.
        :param Sigma: Prediction covariance at locations where the evaluation is done.
        :param dmu_dx: Prediction mean gradient at locations where the evaluation is done.
        :param dSigma_dx: Prediction covariance gradient at locations where the evaluation is done.
        :param y_minimum: The best value evaluated so far by the black box function.
        :param pk: Probabilities of each random variable being smaller than the rest.
        :param symmetric_term: A helper matrix generated by the acquistion function.
        :return: Gradient of the multipoint Expected Improvement where the mean and the covariance are evaluated.
        """
        q, d = Sigma.shape[0], dmu_dx.shape[2]

        # Initialize empty vectors
        grad, L = np.zeros((q, d)), -np.diag(np.ones(q))

        # First sum of the formula
        for k in range(q):
            bk = np.zeros((q, 1))
            bk[k, 0] = y_minimum  # creation of vector b^(k)
            Lk = L.copy()
            Lk[:, k] = 1.0  # linear application to transform Y to Zk
            mk = Lk @ mu[:, None]  # mean of Zk (written m^(k) in the formula)
            Sigk = Lk @ Sigma @ Lk.T  # covariance of Zk
            Sigk = 0.5 * (Sigk + Sigk.T)  # numerical symmetrization

            # Compute the first two terms of the gradient (First row of Equation (6))
            grad_a, mk_dx, Sigk_dx, gradpk, hesspk = self._gradient_of_the_acquisition_first_term(mu, dmu_dx, dSigma_dx,
                                                                                                  y_minimum, pk, k, bk,
                                                                                                  mk, Sigk, Lk)
            grad = grad + grad_a

            # Compute the last three terms of the gradiens (Rows 2-4 of Equation (6))
            grad = grad + self._gradient_of_the_acquisition_second_term(mu, dmu_dx, dSigma_dx, y_minimum,
                                                                        pk, k, bk, mk, Sigk, mk_dx, Sigk_dx,
                                                                        symmetric_term, gradpk, hesspk)

        return grad

    def _get_non_symmetric_and_symmetric_term_k(self, b_k: np.ndarray, mu_k: np.ndarray,
                                                Sigma_k: np.ndarray, pk: np.ndarray, k: int) -> Tuple[np.ndarray, np.ndarray]:
        """
        Helper function for computing Second term in Equation (3) of the paper
        for computing the multipoint Expected Improvement

        :param b_k: Vector b^{k} in Equation (3) in the paper for computing the acquisition function.
        :param mu_k: Mean of Y - Y_min given that Y_k is smaller than any other Y_i, i not k.
        :param Sigma_k: Covariance of Y - Y_min given that Y_k is smaller than any other Y_i, i not k.
        :param pk: Vector of probabilities of Y_k being smaller than the rest for all k.
        :return: Non symmetric and symmetric matrices needed for computing the multipoint Expected Improvement.
        """
        q = mu_k.shape[0]
        non_symmetric_term = np.zeros((q,))
        symmetric_term = np.zeros((q,))
        if self.fast_compute:
            non_symmetric_term[k] = 1. / self.eps * (scipy.stats.multivariate_normal.cdf(b_k - mu_k
                                                                                         + self.eps * Sigma_k[:, k].flatten(),
                                                                                         np.zeros((q,)), Sigma_k) - pk[k])
            symmetric_term[k] = 1.0

        else:
            for i in range(q):
                # First item inside the second sum of Equation (3) in the paper
                non_symmetric_term[i] = Sigma_k[i, k]

                if(i >= k):
                    mik = mu_k[i]
                    sigma_ii_k = Sigma_k[i, i]
                    bik = b_k[i]
                    phi_ik = scipy.stats.norm.pdf(bik, loc=mik, scale=np.sqrt(sigma_ii_k))
                    cik = get_correlations_given_value_of_i(b_k, mu_k, Sigma_k, i).flatten()
                    sigmaik = get_covariance_given_value_of_i(Sigma_k, i)
                    Phi_ik = scipy.stats.multivariate_normal.cdf(cik, np.zeros((q - 1,)), sigmaik)

                    # pdf times cdf in the paper, Equation (3) in the paper, two last items:
                    symmetric_term[i] = phi_ik * Phi_ik
        return symmetric_term, non_symmetric_term

    def _gradient_of_the_acquisition_first_term(self, mu: np.ndarray, dmu_dx: np.ndarray, dSigma_dx: np.ndarray,
                                                y_minimum: float, pk: np.ndarray, k: int, bk: np.ndarray,
                                                mk: np.ndarray, Sigk: np.ndarray, Lk: np.ndarray) -> Tuple[np.ndarray,
                                                np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Helper function for computing the first term of the gradient of the Acquisition (First row of Equation (6))

        :param mu: Prediction mean at locations where the evaluation is done.
        :param dmu_dx: Prediction mean gradient at locations where the evaluation is done.
        :param dSigma_dx: Prediction covariance gradient at locations where the evaluation is done.
        :param y_minimum: The best value evaluated so far by the black box function.
        :param pk: Probabilities of each random variable being smaller than the rest.
        :param k: Index of the assumed winner, as in Y_k < Y_i for all i not k
        :param bk: Vector b^{k} in the Equation (6) in the paper.
        :param mk: Mean of Y - Y_min given that Y_k is smaller than any other Y_i, i not k.
        :param Sigk: Covariance of Y - Y_min given that Y_k is smaller than any other Y_i, i not k.
        :param Lk: Linear application to transform Y to Zk
        :return: Gradient of the first term of the multipoint Expected Improvement,
                 gradient of the mean and covariance of the multivariate normal,
                 and gradient and hessian of the probabilities of variable at k being less than the rest.
        """
        q, d = dSigma_dx.shape[0], dmu_dx.shape[2]
        grad_a = np.zeros((q, d))
        tLk = Lk.T
        Dpk, Sigk_dx, mk_dx = np.zeros((q, d)), np.zeros((q, d, q, q)), np.zeros((q, d, q))

        # term A1: First term in Equation (6) in the paper
        grad_a[k, :] = - dmu_dx[k, k, :] * pk[k]  # q x q x d

        # compute gradient and hessian matrix of the CDF term pk.
        gradpk = Phi_gradient(bk - mk, np.zeros((q, 1)), Sigk)
        hesspk = Phi_hessian(bk, mk, Sigk, gradient=gradpk)
        for l, j in itertools.product(range(q), range(d)):
            Sigk_dx[l, j, :, :] = Lk @ dSigma_dx[:, :, l, j] @ tLk
            mk_dx[l, j, :] = Lk @ dmu_dx[:, l, j]
            Dpk[l, j] = 0.5 * np.sum(hesspk * Sigk_dx[l, j, :, :]) - gradpk[None, :] @ mk_dx[l, j, :, None]

        # term A2: Second term in Equation (6) in the paper
        return grad_a + (y_minimum - mu[k]) * Dpk, mk_dx, Sigk_dx, gradpk, hesspk

    def _gradient_of_the_acquisition_second_term(self, mu: np.ndarray, dmu_dx: np.ndarray, dSigma_dx: np.ndarray,
                                                 y_minimum: float, pk: np.ndarray, k: int, bk: np.ndarray,
                                                 mk: np.ndarray, Sigk: np.ndarray, mk_dx: np.ndarray,
                                                 Sigk_dx: np.ndarray, symmetric_term: np.ndarray,
                                                 gradpk: np.ndarray, hesspk: np.ndarray) -> np.ndarray:
        """
        Helper function for computing the second term of the gradient of the Acquisition (2-4 rows of Equation (6))

        :param mu: Prediction mean at locations where the evaluation is done.
        :param dmu_dx: Prediction mean gradient at locations where the evaluation is done.
        :param dSigma_dx: Prediction covariance gradient at locations where the evaluation is done.
        :param y_minimum: The best value evaluated so far by the black box function.
        :param pk: Probabilities of each random variable being smaller than the rest.
        :param k: Index of the assumed winner, as in Y_k < Y_i for all i not k
        :param bk: Vector b^{k} in the Equation (6) in the paper.
        :param mk: Mean of Y - Y_min given that Y_k is smaller than any other Y_i, i not k.
        :param Sigk: Covariance of Y - Y_min given that Y_k is smaller than any other Y_i, i not k.
        :param mk_dx: Gradient of mk.
        :param Sigk_dx: Gradient of Sigk.
        :param symmetric_term: A helper matrix generated by the acquistion function.
        :param gradpk: Gradient of vector pk.
        :param hesspk: Hessian of vecotr pk.
        :return: Gradient of the first term of the multipoint Expected Improvement.
        """
        q, d = dSigma_dx.shape[0], dmu_dx.shape[2]
        B = np.zeros((q, d))

        # term B (Rows 2-4 in Equation (6) in the paper)
        if self.fast_compute:
            # Numerical approximation through

            gradpk1 = Phi_gradient(bk - mk + Sigk[:, k, None] * self.eps, np.zeros((q, 1)), Sigk)
            hesspk1 = Phi_hessian(bk + Sigk[:, k, None] * self.eps, mk, Sigk, gradient=gradpk1)
            for l, j in itertools.product(range(q), range(d)):
                f1 = (-(mk_dx[None, l, j, :] @ gradpk1[:, None])
                      + self.eps * (Sigk_dx[None, l, j, :, k] @ gradpk1)
                      + 0.5 * np.sum(Sigk_dx[l, j, :, :] * hesspk1))
                f = -(mk_dx[None, l, j, :] @ gradpk[:, None]) + 0.5 * np.sum(Sigk_dx[l, j, :, :] * hesspk)
                B[l, j] = 1.0 / self.eps * (f1 - f)
        else:
            B1, B2, B3 = np.zeros((q, d)), np.zeros((q, d)), np.zeros((q, d))
            for i in range(q):

                # Assign helper variables needed by the gradients (See equation (6) for details)
                ineq = [n for n in range(q) if n is not i]
                Sigk_ik = Sigk[i, k]
                Sigk_ii = Sigk[i, i]
                mk_i = mk[i]
                mk_dx_i = mk_dx[:, :, i]
                bk_i = bk[i, 0]
                ck_pi = (bk[ineq, 0] - mk[ineq, 0] - (bk[i, 0] - mk[i, 0]) / Sigk_ii * Sigk[ineq, :][:, i])[:, None]
                Sigk_pi = 0.5 * (Sigk[ineq, :][:, ineq]
                                 - 1.0 / Sigk_ii * Sigk[ineq, :][:, i, None] @ Sigk[ineq, :][:, i, None].T
                                 + (Sigk[ineq, :][:, ineq]
                                 - 1.0 / Sigk_ii * Sigk[ineq, :][:, i, None] @ Sigk[ineq, :][:, i, None].T).T)
                Sigk_dx_ii = Sigk_dx[:, :, i, i]
                Sigk_dx_ik = Sigk_dx[:, :, i, k]
                phi_ik = np.max([scipy.stats.multivariate_normal.pdf(bk[i, 0], mk_i, Sigk_ii), 1e-11])
                dphi_ik_dSig = ((bk_i - mk_i)**2 / (2.0 * Sigk_ii**2) - 0.5 / Sigk_ii) * phi_ik
                dphi_ik_dm = (bk_i - mk_i) / Sigk_ii * phi_ik
                Phi_ik = symmetric_term[k, i] / phi_ik
                GPhi_ik = Phi_gradient(ck_pi, np.zeros((q - 1, 1)), Sigk_pi)
                HPhi_ik = Phi_hessian(ck_pi, np.zeros((q - 1, 1)), Sigk_pi, GPhi_ik)
                Sigk_mi = Sigk[ineq, i, None]

                # Compute the terms pf the gradient
                for l, j in itertools.product(range(q), range(d)):
                    # B1: Second row of Equation (6) in the paper
                    B1[l, j] = B1[l, j] + Sigk_dx_ik[l, j] * phi_ik * Phi_ik

                    # B2: Third row of Equation (6) in the paper
                    B2[l, j] = B2[l, j] + Sigk_ik * (mk_dx_i[l, j] * dphi_ik_dm
                                                     + dphi_ik_dSig * Sigk_dx_ii[l, j]) * Phi_ik

                    # B3: Fourth row of Equation (6) in the paper
                    dck_pi = (-mk_dx[l, j, ineq] + (mk_dx_i[l, j] * Sigk_ii
                                                    + (bk_i - mk_i) * Sigk_dx_ii[l, j])
                              / (Sigk_ii**2) * Sigk_mi[:, 0]
                              - (bk[i, 0] - mk_i) / Sigk_ii * Sigk_dx[l, j, ineq, i])
                    SigtCross = Sigk_dx[l, j, ineq, i, None] @ Sigk_mi.T
                    dSigk_pi = (Sigk_dx[l, j, :, :][:, ineq][ineq, :]
                                + Sigk_dx_ii[l, j] / (Sigk_ii**2) * (Sigk_mi @ Sigk_mi.T)
                                - (1.0 / Sigk_ii) * (SigtCross + SigtCross.T))
                    B3[l, j] = (B3[l, j] + Sigk_ik * phi_ik * ((GPhi_ik.T @ dck_pi)
                                + 0.5 * np.sum(HPhi_ik * dSigk_pi)))
            B = B1 + B2 + B3
        return B


def get_covariance_given_smallest(Sigma: np.ndarray, k: int) -> np.ndarray:
    """
    Compute covariance of Y - Y_min given that Y_k is smaller than any other Y_i, i not k

    :param Sigma: Covariance of the (not conditioned) multivariate gaussian.
    :param k: Index of random variable.
    :return: Covariance of Y - Y_min given that Y_k is smaller than any other Y_i
    """
    q = Sigma.shape[0]
    res = Sigma.copy()
    neqi = [j for j in range(q) if j is not k]  # Everything else but i
    for i in neqi:
        res[k, i] = Sigma[k, k] - Sigma[k, i]
        res[i, k] = res[k, i]
    for i in neqi:
        for j in [j for j in range(i, q) if j is not k]:
            res[i, j] = Sigma[i, j] + Sigma[k, k] - Sigma[k, i] - Sigma[k, j]
            res[j, i] = res[i, j]
    return res


def get_covariance_given_value_of_i(Sigma: np.ndarray, i: int) -> np.ndarray:
    """
    Covariances of variables k knowing the value of variable i.

    :param Sigma: Covariance of the multivariate Gaussian.
    :param i: Known variable.
    :return: Covariance of Y given that Y_i is known
    """
    result = np.zeros_like(Sigma)
    neqi = [j for j in range(Sigma.shape[0]) if j is not i]  # Everything else but i
    for u, v in itertools.product(neqi, neqi):
        result[u, v] = Sigma[u, v] - Sigma[u, i] * Sigma[v, i] / Sigma[i, i]
    return result[neqi, :][:, neqi]


def get_correlations_given_value_of_i(b: np.ndarray, mu: np.ndarray, Sigma: np.ndarray, i: int) -> np.ndarray:
    """
    Partial correlations of variables k knowing the value of variable i.

    :param b: Known values.
    :param mu: Mean of the multivariate Gaussian knowing i.
    :param Sigma: Covariance of the multivariate gaussian knowing i.
    :param i: known variable.
    :return: Partial correlations between all variables knowing i.
    """
    Sigmai = Sigma[i, :] / Sigma[i, i]
    result = (b - mu) - (b[i] - mu[i]) * Sigmai
    return result[np.arange(Sigma.shape[0]) != i]


def decompose_mvn(x: np.ndarray, mu: np.ndarray, Sigma: np.ndarray, k: list) -> np.ndarray:
    """
    Decompose Multivariate normal to probability of some random variables being smaller than the rest and multiply this
    with the marginal probability distribution of those random variables.

    :param x: Location of decomposition.
    :param mu: Mean of the multivariate Gaussian.
    :param Sigma: Covariance of the multivariate gaussian.
    :param k: Indices of decomposition.
    :return: Weighted probabilities of each variable being smaller than the rest.
    """
    q = Sigma.shape[0]
    if (len(k) == q):
        return scipy.stats.multivariate_normal.pdf(x.flatten(), mu.flatten(), Sigma)
    neqk = [i for i in np.arange(q) if i not in k]
    x1 = x[k, :]
    x2 = x[neqk, :]
    mu1 = mu[k, :]
    Sig22 = Sigma[neqk, :][:, neqk]
    Sig21 = Sigma[neqk, :][:, k]
    Sig12 = Sig21.T
    Sig11 = Sigma[k, :][:, k]
    Sig11Inv = np.linalg.inv(Sig11)
    varcov = Sig22 - Sig21 @ Sig11Inv @ Sig12
    varcov = 0.5 * (varcov + varcov.T)

    if (min(np.diag(varcov)) <= 0):
        varcov[range(len(neqk)), :][:, range(len(neqk))] = np.diag(varcov) * (np.diag(varcov) >= 0) + 1e-9
    moy = mu[neqk] + Sig21 @ Sig11Inv @ (x1 - mu1)
    low = np.minimum(moy, x2) - 5.0 * np.sqrt(np.max(np.abs(varcov)))
    return scipy.stats.multivariate_normal.pdf(x1.flatten(), mu[k].flatten(), Sig11) \
        * scipy.stats.mvn.mvnun(low, x2, moy, varcov, maxpts=1000 * q)[0]


def Phi_gradient(x: np.ndarray, mu: np.ndarray, Sigma: np.ndarray) -> np.ndarray:
    """
    Compute gradient of CDF of multivariate Gaussian distribution.

    :param x: Location where the gradient is evaluated.
    :param mu: Mean of the multivariate Gaussian.
    :param Sigma: Covariance of the multivariate Gaussian.
    :return: Gradient of the CDF of multivariate Gaussian.
    """
    return np.array([decompose_mvn(x, mu, Sigma, [i]) for i in range(Sigma.shape[0])])


def Phi_hessian(x: np.ndarray, mu: np.ndarray, Sigma: np.ndarray, gradient: np.ndarray=None) -> np.ndarray:
    """
    Compute hessian matrix of CDF of multivariate Gaussian distribution.

    :param x: Location where the Hessian is evaluated.
    :param mu: Mean of the multivariate Gaussian.
    :param Sigma: Covariance of the multivariate Gaussian.
    :param gradient: Gradient of the multivariate Gaussian at x.
    :return: Hessian of the CDF of multivariate Gaussian.
    """
    q = Sigma.shape[0]
    if (q == 1):
        res = - (x - mu) / Sigma * scipy.stats.norm.pdf(x, mu, np.sqrt(Sigma))
    else:
        res = np.zeros((q, q))
    for i in range(q - 1):
        for j in range(i + 1, q):
            res[i, j] = decompose_mvn(x, mu, Sigma, [i, j])

    res = res + res.T  # Hessian matrix is symmetric
    # diagonal terms can be computed with the gradient of CDF and the other hessian terms
    if gradient is None:
        res = res - np.diag(((x - mu).flatten() * Phi_gradient(x, mu, Sigma)[:, None]
                             + np.diag(Sigma @ res).flatten()) / np.diag(Sigma).flatten())
    else:
        res = res - np.diag(((x - mu).flatten() * gradient + np.diag(Sigma @ res).flatten()) / np.diag(Sigma).flatten())
    return res



import numpy as np
import scipy.stats as stats


def compute_runtime_feval(trajectory: np.ndarray, target: float) -> int:
    """
    Computes the runtime in terms of number of function evaluations until a performance smaller or
    equal than a specified target value is achieved.

    :param trajectory: trajectory of the performance of an optimizer after each function evaluation
    :param target: target value
    :return: runtime
    """
    rt = np.where(trajectory <= target)[0]
    if len(rt) == 0:
        rt = np.inf
        return rt
    else:
        return rt[0] + 1


def compute_ecdf(error: np.ndarray, targets: np.ndarray) -> tuple:
    """
    Computes the empirical cumulative distribution (ECDF) of the runtime of an optimizer
    across different targets and tasks.

    :param error: matrix with I x R x N entries, where I are the number of instances or tasks,
    R is the number of runs per task and N are the number of function evaluations per task and run
    :param targets: matrix with I x T entries, where I are the number of instances or tasks and T are the number
    of targets values
    :return: ECDF as tuple, where the first entries defines the runtime and the second the CDF
    """
    n_instances = error.shape[0]
    n_runs = error.shape[1]
    n_targets = targets.shape[1]

    runtime = []
    for i in range(n_instances):
        for t in range(n_targets):
            for r in range(n_runs):
                rt = compute_runtime_feval(error[i, r], targets[i, t])
                runtime.append(rt)
    sorted_error = np.sort(runtime)
    yvals = np.arange(len(sorted_error)) / float(len(sorted_error))

    return sorted_error.tolist(), yvals.tolist()


def compute_ranks(errors, n_bootstrap=1000) -> np.ndarray:
    """
    Computes the averaged ranking score in every iteration and for every task..

    :param errors: matrix with M x I x R x N entries, where M are the number of optimizers,
    I are the number of instances or tasks, R is the number of runs per task and
    N are the number of function evaluations per task and run
    :param n_bootstrap: number bootstrap samples to compute the ranks
    :return: the ranks after each iteration as a MxN ndarray, where, as for errors, M are the number of optimizers
    and N are the number of function evaluations
    """
    n_methods = errors.shape[0]
    n_instances = errors.shape[1]
    n_runs = errors.shape[2]
    n_iters = errors.shape[3]

    ranks = np.zeros([n_methods, n_iters])
    for instance_id in range(n_instances):
        for _ in range(n_bootstrap):
            runs = [np.random.randint(n_runs) for i in range(n_methods)]

            rank_samples = [stats.rankdata([errors[i, instance_id, ri, t] for i, ri in enumerate(runs)])
                            for t in range(n_iters)]
            ranks += np.array(rank_samples).T
    ranks /= n_instances * n_bootstrap

    return ranks



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


"""
Contains linear models
"""

import GPy
import numpy as np


class GPyLinearMultiFidelityModel(GPy.core.GP):
    """
    A thin wrapper around GPy.core.GP that does some input checking and provides a default likelihood
    """
    def __init__(self, X: np.ndarray, Y: np.ndarray, kernel: GPy.kern.Kern, n_fidelities: int,
                 likelihood: GPy.likelihoods.Likelihood=None):
        """

        :param X: Training data features with fidelity input appended as last column
        :param Y: Training data targets
        :param kernel: Multi-fidelity kernel
        :param n_fidelities: Number of fidelities in problem
        :param likelihood: GPy likelihood object.
                           Defaults to MixedNoise which has different noise levels for each fidelity
        """

        # Input checks
        if not isinstance(X, np.ndarray):
            raise ValueError('X should be an array')

        if not isinstance(Y, np.ndarray):
            raise ValueError('Y should be an array')

        if X.ndim != 2:
            raise ValueError('X should be 2d')

        if Y.ndim != 2:
            raise ValueError('Y should be 2d')

        if np.any(X[:, -1] >= n_fidelities):
            raise ValueError('One or more points has a higher fidelity index than number of fidelities')

        # Make default likelihood as different noise for each fidelity
        if likelihood is None:
            likelihood = GPy.likelihoods.mixed_noise.MixedNoise(
                [GPy.likelihoods.Gaussian(variance=1.) for _ in range(n_fidelities)])
        y_metadata = {'output_index': X[:, -1].astype(int)}
        super().__init__(X, Y, kernel, likelihood, Y_metadata=y_metadata)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple, Union

from ...core.interfaces.models import IModel
from ...quadrature.interfaces.base_gp import IBaseGaussianProcess
from ...quadrature.kernels.bounds import BoxBounds
from ...quadrature.kernels.integration_measures import IntegrationMeasure


class WarpedBayesianQuadratureModel(IModel):
    """
    The general class for Bayesian quadrature (BQ) with a warped Gaussian process.

    Inference is performed with the warped GP, but the integral is computed on a Gaussian approximation.
    The warping of the base GP is encoded in the methods 'transform' and 'inverse_transform'

    Examples of warping-approximation pairs:
    - a moment matched squared GP (wsabi-m)
    - a linear approximation to a squared GP (wsabi-l)
    - no approximation if there is no warping (Vanilla BQ)
    - ...
    """
    def __init__(self, base_gp: IBaseGaussianProcess, X: np.ndarray, Y: np.ndarray):
        """
        :param base_gp: the underlying GP model
        :param X: the initial locations of integrand evaluations
        :param Y: the values of the integrand at Y
        """
        self.base_gp = base_gp
        # this is to ensure that the base_gp get the correct transform
        self.set_data(X, Y)

    @property
    def X(self) -> np.ndarray:
        return self.base_gp.X

    @property
    def Y(self) -> np.ndarray:
        return self.transform(self.base_gp.Y)

    @property
    def integral_bounds(self) -> Union[None, BoxBounds]:
        return self.base_gp.kern.integral_bounds

    @property
    def reasonable_box_bounds(self) -> BoxBounds:
        return self.base_gp.kern.reasonable_box_bounds

    @property
    def measure(self) -> Union[None, IntegrationMeasure]:
        """probability measure used for integration. Returns None for standard Lebesgue measure (not a probability
        measure) """
        return self.base_gp.kern.measure

    def transform(self, Y: np.ndarray) -> np.ndarray:
        """
        Transform from base-GP to integrand.
        """
        raise NotImplemented

    def inverse_transform(self, Y: np.ndarray) -> np.ndarray:
        """
        Transform from integrand to base-GP.
        """
        raise NotImplemented

    def predict_base(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes predictive means and variances of the warped GP as well as the base GP

        :param X_pred: Locations at which to predict
        :returns: predictive mean and variances of warped GP, and predictive mean and variances of base-GP in that order
        all shapes (n_points, 1).
        """
        raise NotImplemented

    def predict_base_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray,
                                                                             np.ndarray]:
        """
        Computes predictive means and covariance of the warped GP as well as the base GP

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :returns: predictive mean and covariance of warped GP, predictive mean and covariance of base-GP in that order.
        mean shapes both (n_points, 1) and covariance shapes both (n_points, n_points)
        """
        raise NotImplemented

    def predict_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes approximate predictive means and full co-variances of warped GP.

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :return: predictive mean, predictive full covariance of warped-GP, shapes (n_points, 1) and (n_points, n_points)
        """
        return self.predict_base_with_full_covariance(X_pred)[:2]

    def predict(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes predictive means and variances of warped-GP.

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :return: predictive mean, predictive variances of warped-GP, both shapes (n_points, 1)
        """
        return self.predict_base(X_pred)[:2]

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        This method transforms the integrand y values and sets the data
        :param X: observed locations
        :param Y: observed integrand values
        """
        self.base_gp.set_data(X, self.inverse_transform(Y))

    def optimize(self) -> None:
        """Optimizes the hyperparameters of the base GP"""
        self.base_gp.optimize()

    def integrate(self) -> Tuple[float, float]:
        """
        Computes an estimator of the integral as well as its variance.
        :returns: estimator of integral and its variance
        """
        raise NotImplemented



from typing import Tuple

import numpy as np

from ...core.loop.user_function import MultiSourceFunctionWrapper
from ...core import ParameterSpace, ContinuousParameter, InformationSourceParameter


def multi_fidelity_hartmann_3d() -> Tuple[MultiSourceFunctionWrapper, ParameterSpace]:
    r"""
    The function is given by:

    .. math::
        f(x, \alpha) = -\sum_{i=1}^{4} \alpha_i \exp \left( -\sum_{j=1}^{3} A_{i,j}\left( x_j - P_{i, j} \right)^2 \right)

    where

    .. math::
        \mathbf{A} = \begin{bmatrix}
        3.0 & 10 & 30 \\
        0.1 & 10 & 35 \\
        3.0 & 10 & 30 \\
        0.1 & 10 & 35
        \end{bmatrix}

    .. math::
        \mathbf{P} = 10^{-4} \begin{bmatrix}
        3689 & 1170 & 2673 \\
        4699 & 4387 & 7470 \\
        1091 & 8732 & 5547 \\
        381 & 5743 & 8828
        \end{bmatrix}

    The high fidelity function is given by setting:

    .. math::
        \alpha = (1.0, 1.2, 3.0, 3.2)^T

    The middle fidelity is given by setting:

    .. math::
        \alpha = (1.01, 1.19, 2.9, 3.3)^T

    The low fidelity is given by setting:

    .. math::
        \alpha = (1.02, 1.18, 2.8, 3.4)^T

    The domain is given by:

    .. math::
        \mathbf{x}_i \in (0, 1)

    Reference: https://www.sfu.ca/~ssurjano/hart3.html

    :return: Tuple of MultiSourceFunctionWrapper and ParameterSpace
    """
    A = np.array([[3, 10, 30], [0.1, 10, 35], [3, 10, 30], [.1, 10, 35]])
    P = 1e-4 * np.array([[3689, 1170, 2673], [4699, 4387, 7470], [1091, 8732, 5547], [381, 5743, 8828]])

    alpha = np.array([1.0, 1.2, 3.0, 3.2])
    delta = np.array([0.01, -0.01, -0.1, 0.1])

    def high(x):

        res = 0
        for i in range(4):
            temp = 0
            for j in range(3):
                temp -= A[i][j] * np.power(x[:, j] - P[i][j], 2)
            res += alpha[i] * np.exp(temp)

        return res[:, None]

    def medium(x):

        alpha_m = alpha + delta

        res = 0
        for i in range(4):
            temp = 0
            for j in range(3):
                temp -= A[i][j] * np.power(x[:, j] - P[i][j], 2)
            res += alpha_m[i] * np.exp(temp)

        return res[:, None]

    def low(x):

        alpha_l = alpha + 2 * delta

        res = 0
        for i in range(4):
            temp = 0
            for j in range(3):
                temp -= A[i][j] * np.power(x[:, j] - P[i][j], 2)
            res += alpha_l[i] * np.exp(temp)

        return res[:, None]

    space = ParameterSpace([ContinuousParameter('x1', 0., 1.), ContinuousParameter('x2', 0., 1.),
                            ContinuousParameter('x3', 0., 1.), InformationSourceParameter(3)])

    fcn_wrapper = MultiSourceFunctionWrapper([low, medium, high])

    return fcn_wrapper, space



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import abc
from typing import Tuple

import numpy as np


class Acquisition(abc.ABC):
    """ Acquisition base class """
    def __add__(self, other: 'Acquisition') -> 'Sum':
        """
        Overloads self + other
        """
        # If both acquisitions implement gradients, the gradients can be available in the sum
        return Sum(self, other)

    def __mul__(self, other: 'Acquisition') -> 'Product':
        """
        Overloads self * other
        """
        return Product(self, other)

    def __rmul__(self, other: 'Acquisition') -> 'Product':
        """
        Overloads other * self
        """
        return Product(other, self)

    def __truediv__(self, denominator: 'Acquisition') -> 'Quotient':
        """
        Overloads self / other
        """
        return Quotient(self, denominator)

    @abc.abstractmethod
    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Abstract method. Evaluates the acquisition function.

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values
        :return: (n_points x 1) array of acquisition function values
        """
        pass

    @property
    @abc.abstractmethod
    def has_gradients(self) -> bool:
        """
        Abstract property. Whether acquisition value has analytical gradient calculation available.

        :return: True if gradients are available
        """
        pass

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Optional abstract method that must be implemented if has_gradients returns True.
        Evaluates value and gradient of acquisition function at x.

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values and gradient
        :return: Tuple contains an (n_points x 1) array of acquisition function values and (n_points x n_dims) array of
                 acquisition function gradients with respect to x
        """
        raise NotImplementedError("Gradients not implemented for this acquisition function")

    def update_parameters(self) -> None:
        """
        Performs any updates to parameters that needs to be done once per outer loop iteration
        """
        pass


class Quotient(Acquisition):
    """
    Acquisition for division of two acquisition functions
    """
    def __init__(self, numerator: Acquisition, denominator: Acquisition):
        """

        :param numerator: Acquisition function to act as numerator in quotient
        :param denominator: Acquisition function to act as denominator in quotient
        """
        self.numerator = numerator
        self.denominator = denominator

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate division of the two acquisition functions

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values
        :return: (n_points x 1) array of acquisition function values
        """
        return self.numerator.evaluate(x) / self.denominator.evaluate(x)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluate value and gradient of acquisition function at x

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values and gradient
        :return: Tuple contains an (n_points x 1) array of acquisition function values and (n_points x n_dims) array of
                 acquisition function gradients with respect to x
        """
        # Evaluate both acquisition functions
        numerator_value, numerator_gradients = self.numerator.evaluate_with_gradients(x)
        denominator_value, denominator_gradients = self.denominator.evaluate_with_gradients(x)

        value = numerator_value / denominator_value
        # Calculate gradient of acquisition
        gradient = (numerator_gradients / denominator_value) \
                   - ((denominator_gradients * numerator_value) / (denominator_value**2))
        return value, gradient

    @property
    def has_gradients(self) -> bool:
        """
        Whether acquisition value has analytical gradient calculation available.

        :return: True if gradients are available
        """
        return self.denominator.has_gradients and self.numerator.has_gradients

    def update_parameters(self) -> None:
        """
        Performs any updates to parameters that needs to be done once per outer loop iteration
        """
        self.denominator.update_parameters()
        self.numerator.update_parameters()


class Product(Acquisition):
    """
    Acquisition for product of two or more acquisition functions
    """
    def __init__(self, acquisition_1: Acquisition, acquisition_2: Acquisition):
        """

        :param acquisition_1: Acquisition function in product
        :param acquisition_2: Other acquisition function in product
        """
        self.acquisition_1 = acquisition_1
        self.acquisition_2 = acquisition_2

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate product of the two acquisition functions

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values
        :return: (n_points x 1) array of acquisition function values
        """
        return self.acquisition_1.evaluate(x) * self.acquisition_2.evaluate(x)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluate value and gradient of acquisition function at x

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values and gradient
        :return: Tuple contains an (n_points x 1) array of acquisition function values and (n_points x n_dims) array of
                 acquisition function gradients with respect to x
        """
        # Evaluate acquisitions
        value_1, grad_1 = self.acquisition_1.evaluate_with_gradients(x)
        value_2, grad_2 = self.acquisition_2.evaluate_with_gradients(x)

        # Calculate gradient
        grad_total = value_1 * grad_2 + value_2 * grad_1

        return value_1 * value_2, grad_total

    @property
    def has_gradients(self):
        """
        Whether acquisition value has analytical gradient calculation available.

        :return: True if gradients are available
        """
        return self.acquisition_1.has_gradients and self.acquisition_2.has_gradients

    def update_parameters(self) -> None:
        """
        Performs any updates to parameters that needs to be done once per outer loop iteration
        """
        self.acquisition_1.update_parameters()
        self.acquisition_2.update_parameters()


class Sum(Acquisition):
    """
    Acquisition for sum of two acquisition functions
    """

    def __init__(self, acquisition_1: Acquisition, acquisition_2: Acquisition) -> None:
        """
        :param acquisition_1: An acquisition function in sum
        :param acquisition_2: Other acquisition function in sum
        """
        self.acquisition_1 = acquisition_1
        self.acquisition_2 = acquisition_2

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate sum of the two acquisition functions

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values
        :return: (n_points x 1) array of acquisition function values
        """
        return self.acquisition_1.evaluate(x) + self.acquisition_2.evaluate(x)

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluate value and gradient of acquisition function at x

        :param x: (n_points x n_dims) array of points at which to calculate acquisition function values and gradient
        :return: Tuple contains an (n_points x 1) array of acquisition function values and (n_points x n_dims) array of
                 acquisition function gradients with respect to x
        """
        # Evaluate first acquisition with gradients
        value_1, grad_1 = self.acquisition_1.evaluate_with_gradients(x)
        value_2, grad_2 = self.acquisition_2.evaluate_with_gradients(x)

        return value_1 + value_2, grad_1 + grad_2

    @property
    def has_gradients(self):
        """
        Whether acquisition value has analytical gradient calculation available.

        :return: True if gradients are available
        """
        return self.acquisition_1.has_gradients and self.acquisition_2.has_gradients

    def update_parameters(self) -> None:
        """
        Performs any updates to parameters that needs to be done once per outer loop iteration
        """
        self.acquisition_1.update_parameters()
        self.acquisition_2.update_parameters()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from scipy.linalg import lapack
from typing import Tuple

from ...quadrature.interfaces.base_gp import IBaseGaussianProcess
from ...core.interfaces.models import IDifferentiable
from .warped_bq_model import WarpedBayesianQuadratureModel


class VanillaBayesianQuadrature(WarpedBayesianQuadratureModel, IDifferentiable):
    """
    Vanilla Bayesian quadrature.

    Vanilla Bayesian quadrature uses a Gaussian process as surrogate for the integrand.
    """

    def __init__(self, base_gp: IBaseGaussianProcess, X: np.ndarray, Y: np.ndarray):
        """
        :param base_gp: a model derived from BaseGaussianProcess
        :param X: the initial locations of integrand evaluations
        :param Y: the values of the integrand at Y
        """
        super(VanillaBayesianQuadrature, self).__init__(base_gp=base_gp, X=X, Y=Y)

    def transform(self, Y: np.ndarray) -> np.ndarray:
        """ Transform from base-GP to integrand """
        return Y

    def inverse_transform(self, Y: np.ndarray) -> np.ndarray:
        """ Transform from integrand to base-GP """
        return Y

    def predict_base(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Computes predictive means and variances of the warped GP as well as the base GP

        :param X_pred: Locations at which to predict
        :returns: predictive mean and variances of warped GP, and predictive mean and variances of base-GP in that order
        all shapes (n_points, 1).
        """
        m, cov = self.base_gp.predict(X_pred)
        return m, cov, m, cov

    def predict_base_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray,
                                                                             np.ndarray]:
        """
        Computes predictive means and covariance of the warped GP as well as the base GP

        :param X_pred: Locations at which to predict, shape (n_points, input_dim)
        :returns: predictive mean and covariance of warped GP, predictive mean and covariance of base-GP in that order.
        mean shapes both (n_points, 1) and covariance shapes both (n_points, n_points)
        """
        m, cov = self.base_gp.predict_with_full_covariance(X_pred)
        return m, cov, m, cov

    def integrate(self) -> Tuple[float, float]:
        """
        Computes an estimator of the integral as well as its variance.

        :returns: estimator of integral and its variance
        """
        kernel_mean_X = self.base_gp.kern.qK(self.X)
        integral_mean = np.dot(kernel_mean_X, self.base_gp.graminv_residual())[0, 0]
        integral_var = self.base_gp.kern.qKq() - (kernel_mean_X @ self.base_gp.solve_linear(kernel_mean_X.T))[0, 0]
        return integral_mean, integral_var

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple:
        """
        Computes and returns model gradients of mean and variance at given points

        :param X: points to compute gradients at, shape (num_points, dim)
        :returns: Tuple of gradients of mean and variance, shapes of both (num_points, dim)
        """
        # gradient of mean
        d_mean_dx = (self.base_gp.kern.dK_dx1(X, self.X) @ self.base_gp.graminv_residual())[:, :, 0].T

        # gradient of variance
        dKdiag_dx = self.base_gp.kern.dKdiag_dx(X)
        dKxX_dx1 = self.base_gp.kern.dK_dx1(X, self.X)
        graminv_KXx = self.base_gp.solve_linear(self.base_gp.kern.K(self.base_gp.X, X))
        d_var_dx = dKdiag_dx - 2. * (dKxX_dx1 * np.transpose(graminv_KXx)).sum(axis=2, keepdims=False)

        return d_mean_dx, d_var_dx.T



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import abc
import logging
from typing import Dict, Optional, Tuple

import numpy as np

from .. import ParameterSpace
from ..acquisition import Acquisition
from .context_manager import Context, ContextManager

_log = logging.getLogger(__name__)


class AcquisitionOptimizerBase(abc.ABC):
    """
    Base class for acquisition optimizers
    """
    def __init__(self, space: ParameterSpace):
        """
        :param space: Parameter space containing entire input domain including any context variables
        """
        self.space = space

    def _validate_context_parameters(self, context: Dict[str, any]):
        for context_name, context_value in context.items():
            # Check parameter exists in space
            if context_name not in self.space.parameter_names:
                raise ValueError(context_name + ' appears as variable in context but not in the parameter space.')

            # Log warning if context parameter is out of domain
            param = self.space.get_parameter_by_name(context_name)
            if param.check_in_domain(context_value) is False:
                _log.warning(context_name + ' with value ' + str(context_value), ' is out of the domain')
            else:
                _log.info('Parameter ' + context_name + ' fixed to ' + str(context_value))

    @abc.abstractmethod
    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager) -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of optimization. See class docstring for details.

        :param acquisition: The acquisition function to be optimized
        :param context_manager: Optimization context manager.
        :return: Tuple of (location of maximum, acquisition value at maximizer)
        """
        pass

    def optimize(self, acquisition: Acquisition, context: Optional[Context] = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Optimizes the acquisition function.

        :param acquisition: The acquisition function to be optimized
        :param context: Optimization context.
                        Determines whether any variable values should be fixed during the optimization
        :return: Tuple of (location of maximum, acquisition value at maximizer)
        """
        if context is None:
            context = dict()
        else:
            self._validate_context_parameters(context)
        context_manager = ContextManager(self.space, context)
        max_x, max_value = self._optimize(acquisition, context_manager)

        # Optimization might not match any encoding exactly
        # Rounding operation here finds the closest encoding
        rounded_max_x = self.space.round(max_x)

        if not np.array_equal(max_x, rounded_max_x):
            # re-evaluate if x changed while rounding to make sure value is correct
            return rounded_max_x, acquisition.evaluate(rounded_max_x)
        else:
            return max_x, max_value



from typing import Union, Callable, Tuple

import numpy as np

from emukit.core.acquisition import Acquisition
from emukit.core.interfaces import IModel, IPriorHyperparameters


class IntegratedHyperParameterAcquisition(Acquisition):
    """
    This acquisition class provides functionality for integrating any acquisition function over model hyper-parameters
    """
    def __init__(self, model: Union[IModel, IPriorHyperparameters], acquisition_generator: Callable, n_samples: int=10,
                 n_burnin: int=100, subsample_interval: int=10, step_size: float=1e-1, leapfrog_steps: int=20):
        """
        :param model: An emukit model that implements IPriorHyperparameters
        :param acquisition_generator: Function that returns acquisition object when given the model as the only argument
        :param n_samples: Number of hyper-parameter samples
        :param n_burnin: Number of initial samples not used.
        :param subsample_interval: Interval of subsampling from HMC samples.
        :param step_size: Size of the gradient steps in the HMC sampler.
        :param leapfrog_steps: Number of gradient steps before each Metropolis Hasting step.
        """
        self.model = model
        self.acquisition_generator = acquisition_generator
        self.n_samples = n_samples
        self.samples = self.model.generate_hyperparameters_samples(n_samples, n_burnin,
                                                                   subsample_interval, step_size, leapfrog_steps)

        acquisition = self.acquisition_generator(model)
        self._has_gradients = acquisition.has_gradients

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluate acquisition by integrating over the hyper-parameters of the model
        :param x: locations where the evaluation is done.
        :return: Array with integrated acquisition value at all input locations
        """
        acquisition_value = 0
        for sample in self.samples:
            self.model.fix_model_hyperparameters(sample)
            acquisition = self.acquisition_generator(self.model)
            acquisition_value += acquisition.evaluate(x)

        return acquisition_value / self.n_samples

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple:
        """
        Computes the acquisition value and its derivative integrating over the hyper-parameters of the model

        :param x: locations where the evaluation with gradients is done.
        :return: tuple containing the integrated expected improvement at the points x and its gradient.
        """

        if x.ndim == 1:
            x = x[None, :]

        acquisition_value = 0
        d_acquisition_dx = 0

        for sample in self.samples:
            self.model.fix_model_hyperparameters(sample)
            acquisition = self.acquisition_generator(self.model)
            improvement_sample, d_improvement_dx_sample = acquisition.evaluate_with_gradients(x)
            acquisition_value += improvement_sample
            d_acquisition_dx += d_improvement_dx_sample

        return acquisition_value / self.n_samples, d_acquisition_dx / self.n_samples

    def update_parameters(self):
        self.samples = self.model.generate_hyperparameters_samples(self.n_samples)

    @property
    def has_gradients(self) -> bool:
        """Returns that this acquisition has gradients"""
        return self._has_gradients



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from . import interfaces  # noqa: F401
from . import kernels  # noqa: F401
from . import loop  # noqa: F401
from . import methods  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, Optional

import numpy as np

from .. import InformationSourceParameter, ParameterSpace
from ..acquisition import Acquisition
from .acquisition_optimizer import AcquisitionOptimizerBase
from .context_manager import ContextManager, Context


class MultiSourceAcquisitionOptimizer(AcquisitionOptimizerBase):
    """
    Optimizes the acquisition function by finding the optimum input location at each information source, then picking
    the information source where the value of the acquisition at the optimum input location is highest.
    """
    def __init__(self, acquisition_optimizer: AcquisitionOptimizerBase, space: ParameterSpace) -> None:
        """
        :param acquisition_optimizer: Optimizer to use for optimizing the acquisition once the information source
                                      has been fixed
        :param space: Domain to search for maximum over
        """
        super().__init__(space)
        self.acquisition_optimizer = acquisition_optimizer
        self.source_parameter = self._get_information_source_parameter()
        self.n_sources = np.array(self.source_parameter.domain).size

    def _get_information_source_parameter(self) -> InformationSourceParameter:
        """
        :return: The parameter containing the index of the information source
        """
        source_parameter = [param for param in self.space.parameters if isinstance(param, InformationSourceParameter)]
        if len(source_parameter) == 0:
            raise ValueError('No source parameter found')
        return source_parameter[0]

    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager) -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of abstract method. Does nothing, optimize overwritten instead.

        See AcquisitionOptimizerBase._optimizer for parameter descriptions.
        See class docstring for implementation details.
        """
        pass

    def optimize(self, acquisition: Acquisition, context: Optional[Context] = None) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the location and source of the next point to evaluate by finding the maximum input location at each
        information source, then picking the information source where the value of the acquisition at the optimum input
        location is highest.

        :param acquisition: The acquisition function to be optimized
        :param context: Contains variables to fix through optimization of acquisition function. The dictionary key is
                        the parameter name and the value is the value to fix the parameter to.
        :return: A tuple of (location of maximum, acquisition value at maximum)
        """
        f_maxs = np.zeros((len(self.source_parameter.domain)))
        x_opts = []

        if context is None:
            context = dict()
        elif self.source_parameter.name in context:
            # Information source parameter already has a context so just optimise the acquisition at this source
            return self.acquisition_optimizer.optimize(acquisition, context)

        # Optimize acquisition for each information source
        for i in range(len(self.source_parameter.domain)):
            # Fix the source using a dictionary, the key is the name of the parameter to fix and the value is the
            # value to which the parameter is fixed
            context[self.source_parameter.name] = self.source_parameter.domain[i]
            x, f_maxs[i] = self.acquisition_optimizer.optimize(acquisition, context)
            x_opts.append(x)
        best_source = np.argmax(f_maxs)
        return x_opts[best_source], np.max(f_maxs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple

from ...core.interfaces import IModel
from ..kernels.quadrature_kernels import QuadratureKernel


class IBaseGaussianProcess(IModel):
    """
    Interface for the quadrature base-GP model
    An instance of this can be passed as 'base_gp' to an ApproximateWarpedGPSurrogate object.

    If this GP is initialized with data, use the raw evaluations Y of the integrand and not transformed values.
    """

    def __init__(self, kern: QuadratureKernel) -> None:
        """
        If this GP is initialized with data X, Y, use the raw evaluations Y of the integrand and not transformed values
        as this is a general class that can be used with various quadrature methods. The transformation will be
        performed automatically when the quadrature method is initialized subsequently.
        :param kern: a quadrature kernel
        """
        self.kern = kern

    @property
    def observation_noise_variance(self) -> np.float:
        """
        Gaussian observation noise variance
        :return: The noise variance from some external GP model
        """
        raise NotImplementedError

    def predict_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predictive mean and full co-variance at the locations X_pred

        :param X_pred: points at which to predict, with shape (num_points, input_dim)
        :return: Predictive mean, predictive full co-variance shapes (num_points, num_points)
        """
        raise NotImplementedError

    def solve_linear(self, z: np.ndarray) -> np.ndarray:
        """
        Solve the linear system G(X, X)x=z for x.
        G(X, X) is the Gram matrix :math:`G(X, X) = K(X, X) + \sigma^2 I`, of shape (num_dat, num_dat) and z is a
        matrix of shape (num_dat, num_obs).

        :param z: a matrix of shape (num_dat, num_obs)
        :return: the solution to the linear system G(X, X)x = z, shape (num_dat, num_obs)
        """
        raise NotImplementedError

    def graminv_residual(self) -> np.ndarray:
        """
        The inverse Gram matrix multiplied with the mean-corrected data

        ..math::

            (K_{XX} + \sigma^2 I)^{-1} (Y - m(X))

        where the data is given by {X, Y} and m is the prior mean and sigma^2 the observation noise

        :return: the inverse Gram matrix multiplied with the mean-corrected data with shape: (number of datapoints, 1)
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Iterable, Union, Tuple, List

import numpy as np

from .parameter import Parameter


class DiscreteParameter(Parameter):
    """
    A parameter that takes a discrete set of values where the order and spacing of values is important
    """
    def __init__(self, name: str, domain: Iterable):
        """
        :param name: Name of parameter
        :param domain: valid values the parameter can have
        """
        self.name = name
        self.domain = domain

    def check_in_domain(self, x: Union[np.ndarray, Iterable, float]) -> bool:
        """
        Checks if the points in x are in the set of allowed values

        :param x: 1d numpy array of points to check
        :param x:    1d numpy array of points to check
                  or 2d numpy array with shape (n_points, 1) of points to check
                  or Iterable of points to check
                  or float of single point to check
        :return: A boolean indicating whether each point is in domain
        """
        if np.isscalar(x):
            x = [x]
        elif isinstance(x, np.ndarray):
            if x.ndim == 2 and x.shape[1] == 1:
                x = x.ravel()
            elif x.ndim > 1:
                raise ValueError("Expected x shape (n,) or (n, 1), actual is {}".format(x.shape))
        return set(x).issubset(set(self.domain))

    @property
    def bounds(self) -> List[Tuple]:
        """
        Returns a list containing one tuple of min and max values parameter can take
        """
        return [(min(self.domain), max(self.domain))]

    def round(self, x: np.ndarray) -> np.ndarray:
        """
        Rounds each row in x to represent a valid value for this discrete variable

        :param x: A 2d array Nx1 to be rounded
        :returns: An array Nx1 where each row represents a value from the domain
                  that is closest to the corresponding row in x
        """
        if x.ndim != 2:
            raise ValueError("Expected 2d array, got " + str(x.ndim))

        if x.shape[1] != 1:
            raise ValueError("Expected single column array, got {}".format(x.shape[1]))

        x_rounded = []
        for row in x:
            value = row[0]
            rounded_value = min(self.domain, key=lambda d: abs(d - value))
            x_rounded.append([rounded_value])

        return np.row_stack(x_rounded)

    def sample_uniform(self, point_count: int) -> np.ndarray:
        """
        Generates multiple uniformly distributed random parameter points.

        :param point_count: number of data points to generate.
        :returns: Generated points with shape (point_count, num_features)
        """
        indices = np.random.randint(0, len(self.domain), point_count)
        return np.asarray(self.domain)[indices, None]


class InformationSourceParameter(DiscreteParameter):
    def __init__(self, n_sources: int) -> None:
        """
        :param n_sources: Number of information sources in the problem
        """
        super().__init__('source', list(range(n_sources)))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple

from ...core.acquisition import Acquisition
from ..methods import VanillaBayesianQuadrature


class SquaredCorrelation(Acquisition):
    """
    This acquisition function is the correlation between the integral and the new point(s) under a GP-model.

    SquaredCorrelation is identical to the integral-variance-reduction acquisition up to a global normalizing constant!

    .. math::
        \rho^2(x) = \frac{(\int k_N(x_1, x)\mathrm{d}x_1)^2}{\mathfrac{v}_N v_N(x)}\in [0, 1]

    where :math:`\mathfrac{v}_N` is the current integral variance given N observations X, :math:`v_N(x)` is the
    predictive integral variance if point x was added newly, and :math:`k_N(x_1, x)` is the posterior kernel function.
    """

    def __init__(self, model: VanillaBayesianQuadrature):
        """
        :param model: The vanilla Bayesian quadrature model
        """
        self.model = model

    def has_gradients(self) -> bool:
        return True

    def evaluate(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the acquisition function at x.

        :param x: (n_points x input_dim) locations where to evaluate
        :return: (n_points x 1) the acquisition function value at x
        """
        return self._evaluate(x)[0]

    def _evaluate(self, x: np.ndarray) -> Tuple[np.ndarray, np.float, np.ndarray, np.ndarray]:
        """
        Evaluates the acquisition function at x.

        :param x: (n_points x input_dim) locations where to evaluate
        :return: the acquisition function value at x, shape (n_points x 1), current integral variance,
        predictive variance + noise, predictive covariance between integral and x, shapes of the latter
        two (n_points, 1).
        """
        integral_current_var, y_predictive_var, predictive_cov = self._value_terms(x)
        squared_correlation = predictive_cov**2 / (integral_current_var * y_predictive_var)
        return squared_correlation, integral_current_var, y_predictive_var, predictive_cov

    def evaluate_with_gradients(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Evaluate the acquisition function with gradient

        :param x: (n_points x input_dim) locations where to evaluate
        :return: acquisition value and corresponding gradient at x, shapes (n_points, 1) and (n_points, input_dim)
        """
        # value
        squared_correlation, integral_current_var, y_predictive_var, predictive_cov = self._evaluate(x)

        # gradient
        d_y_predictive_var_dx, d_predictive_cov_dx = self._gradient_terms(x)
        first_term = 2. * predictive_cov * d_predictive_cov_dx
        second_term = (predictive_cov**2 / y_predictive_var) * d_y_predictive_var_dx
        normalization = integral_current_var * y_predictive_var
        squared_correlation_gradient = (first_term - second_term) / normalization

        return squared_correlation, squared_correlation_gradient

    def _value_terms(self, x: np.ndarray) -> Tuple[np.float, np.ndarray, np.ndarray]:
        """
        computes the terms needed for the squared correlation

        :param x: (n_points x input_dim) locations where to evaluate
        :return: current integral variance, predictive variance + noise, predictive covariance between integral and x,
           shapes of the latter two arrays are (n_points, 1).
        """
        integral_current_var = self.model.integrate()[1]
        y_predictive_var = self.model.predict(x)[1] + self.model.base_gp.observation_noise_variance

        qKx = self.model.base_gp.kern.qK(x)
        qKX = self.model.base_gp.kern.qK(self.model.base_gp.X)

        graminv_KXx = self.model.base_gp.solve_linear(self.model.base_gp.kern.K(self.model.base_gp.X, x))
        predictive_cov = np.transpose(qKx - np.dot(qKX, graminv_KXx))
        return integral_current_var, y_predictive_var, predictive_cov

    def _gradient_terms(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Computes the terms needed for the gradient of the squared correlation

        :param x: (n_points x input_dim) locations where to evaluate
        :return: the gradient of (y_predictive_var, predictive_cov) wrt. x at param x, shapes (n_points, input_dim)
        """
        # gradient of predictive variance of y
        d_y_predictive_var_dx = self.model.get_prediction_gradients(x)[1].T

        # gradient of predictive covariance between integral and (x, y)-pair
        dqKx_dx = np.transpose(self.model.base_gp.kern.dqK_dx(x))

        qKX = self.model.base_gp.kern.qK(self.model.base_gp.X)
        qKX_graminv = np.transpose(self.model.base_gp.solve_linear(qKX.T))  # (1, N)

        dKXx_dx2 = self.model.base_gp.kern.dK_dx2(self.model.X, x)
        d_predictive_cov_dx = dqKx_dx - np.dot(qKX_graminv, np.transpose(dKXx_dx2))[0, :, :]

        return np.transpose(d_y_predictive_var_dx), d_predictive_cov_dx



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .base_gp import IBaseGaussianProcess  # noqa: F401
from .standard_kernels import IRBF, IStandardKernel  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from ...core.loop.loop_state import create_loop_state
from ...core.loop import OuterLoop, SequentialPointCalculator, FixedIntervalUpdater, ModelUpdater
from ...core.optimization import AcquisitionOptimizerBase, GradientAcquisitionOptimizer
from ...core.parameter_space import ParameterSpace
from ...core.acquisition import Acquisition
from ..methods import VanillaBayesianQuadrature
from ..acquisitions import IntegralVarianceReduction


class VanillaBayesianQuadratureLoop(OuterLoop):
    def __init__(self, model: VanillaBayesianQuadrature, acquisition: Acquisition = None,
                 model_updater: ModelUpdater = None, acquisition_optimizer: AcquisitionOptimizerBase = None):
        """
        The loop for vanilla Bayesian Quadrature

        :param model: the vanilla Bayesian quadrature method
        :param acquisition: The acquisition function that is used to collect new points.
        default, IntegralVarianceReduction
        :param model_updater: Defines how and when the quadrature model is updated if new data arrives.
                              Defaults to updating hyper-parameters every iteration.
        :param acquisition_optimizer: Optimizer selecting next evaluation points by maximizing acquisition.
                                      Gradient based optimizer is used if None. Defaults to None.
        """

        if acquisition is None:
            acquisition = IntegralVarianceReduction(model)

        if model_updater is None:
            model_updater = FixedIntervalUpdater(model, 1)

        space = ParameterSpace(model.reasonable_box_bounds.convert_to_list_of_continuous_parameters())
        if acquisition_optimizer is None:
            acquisition_optimizer = GradientAcquisitionOptimizer(space)
        candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)
        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updater, loop_state)

        self.model = model



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .vanilla_bq_loop import VanillaBayesianQuadratureLoop  # noqa: F401
from .bayesian_monte_carlo_loop import BayesianMonteCarlo  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from typing import Tuple, List
import numpy as np

from emukit.core.loop.user_function import UserFunctionWrapper


def circular_gaussian(mean: np.float = 0., variance: np.float = 1.) -> Tuple[UserFunctionWrapper,
                                                                             List[Tuple[float, float]]]:
    """
    2D toy integrand that is a Gaussian on a circle

    .. math::

        (2\pi \sigma^2)^{-0.5} r^2 e^{-\frac{(r - \mu)^2}{2 \sigma^2}}

    :param mean: the mean of the circular Gaussian in units of radius (must be >= 0, defaults to 0)
    :param variance: variance of the Gaussian (must be > 0, defaults to 1.)
    :return: the wrapped test function, and the integrals bounds (defaults to area [-3, 3]^2).
    """
    func = lambda x: _circular_gaussian(x, mean, variance)
    integral_bounds = 2 * [(-3., 3.)]
    return UserFunctionWrapper(func), integral_bounds


def _circular_gaussian(x: np.ndarray, mean: np.float, variance: np.float) -> np.ndarray:
    """
    :param x: locations for evaluation (num_points, 2)
    :param mean: mean of Gaussian in units of radius (must be >= 0)
    :param variance: variance of the Gaussian (must be > 0)
    :return: the function values at x, shape (num_points, 1)
    """
    norm_x = np.sqrt((x ** 2).sum(axis=1, keepdims=True))
    return norm_x**2 * np.exp(- (norm_x - mean)**2 / (2. * variance)) / np.sqrt(2. * np.pi * variance)



from .borehole import multi_fidelity_borehole_function  # noqa: F401
from .branin import multi_fidelity_branin_function  # noqa: F401
from .currin import multi_fidelity_currin_function  # noqa: F401
from .hartmann import multi_fidelity_hartmann_3d  # noqa: F401
from .park import multi_fidelity_park_function  # noqa: F401



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
try:
    import pyDOE
except ImportError:
    raise ImportError('pyDOE needs to be installed in order to use latin design')

from .base import ModelFreeDesignBase
from .. import ParameterSpace


class LatinDesign(ModelFreeDesignBase):
    """
    Latin hypercube experiment design.

    Based on pyDOE implementation. For further reference see https://pythonhosted.org/pyDOE/randomized.html#latin-hypercube
    """
    def __init__(self, parameter_space: ParameterSpace) -> None:
        """
        :param parameter_space: The parameter space to generate design for.
        """
        super(LatinDesign, self).__init__(parameter_space)

    def get_samples(self, point_count: int) -> np.ndarray:
        """
        Generates requested amount of points.

        :param point_count: Number of points required.
        :return: A numpy array with shape (point_count x space_dim)
        """
        bounds = self.parameter_space.get_bounds()
        X_design_aux = pyDOE.lhs(len(bounds), point_count, criterion='center')
        ones = np.ones((X_design_aux.shape[0], 1))

        lower_bound = np.asarray(bounds)[:, 0].reshape(1, len(bounds))
        upper_bound = np.asarray(bounds)[:, 1].reshape(1, len(bounds))
        diff = upper_bound - lower_bound

        X_design = np.dot(ones, lower_bound) + X_design_aux * np.dot(ones, diff)

        samples = self.parameter_space.round(X_design)

        return samples



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

import logging
from typing import Tuple

import numpy as np

from .acquisition_optimizer import AcquisitionOptimizerBase
from .context_manager import ContextManager
from .. import ParameterSpace
from ..acquisition import Acquisition

_log = logging.getLogger(__name__)


class RandomSearchAcquisitionOptimizer(AcquisitionOptimizerBase):
    """ Optimizes the acquisition function by evaluating at random points.
    Can be used for discrete and continuous acquisition functions.
    """
    def __init__(self, space: ParameterSpace, num_eval_points: int) -> None:
        """
        :param space: The parameter space spanning the search problem.
        :param num_eval_points: Number of random sampled points which are evaluated per optimization.
        """
        super().__init__(space)
        self.num_eval_points = num_eval_points

    def _optimize(self, acquisition: Acquisition, context_manager: ContextManager)\
        -> Tuple[np.ndarray, np.ndarray]:
        """
        Implementation of abstract method.

        See AcquisitionOptimizerBase._optimizer for parameter descriptions.
        See class docstring for implementation details.
        """
        _log.info("Starting random search optimization of acquisition function {}"
                  .format(type(acquisition)))
        samples_contextfree = context_manager.contextfree_space.sample_uniform(
            self.num_eval_points)
        samples = context_manager.expand_vector(samples_contextfree)
        acquisition_values = acquisition.evaluate(samples)
        max_index = np.argmax(acquisition_values)

        return samples[[max_index]], acquisition_values[[max_index]]



from typing import List

import numpy as np


class BenchmarkResult:
    def __init__(self, loop_names: List[str], n_repeats: int, metric_names: List[str]):
        """

        :param loop_names: List of loop names
        :param n_repeats: Number of random restarts in benchmarking
        :param metric_names: List of metric names
        """

        self.loop_names = loop_names
        self.n_repeats = n_repeats
        self.metric_names = metric_names

        self._results = dict()
        for loop_name in loop_names:
            self._results[loop_name] = dict()
            for metric_name in metric_names:
                self._results[loop_name][metric_name] = []
                for i in range(n_repeats):
                    self._results[loop_name][metric_name].append([])

    def add_results(self, loop_name: str, i_repeat: int, metric_name: str, metric_values: np.ndarray) -> None:
        """
        Add results for a specific loop, metric and repeat combination

        :param loop_name: Name of loop
        :param i_repeat: Index of repeat
        :param metric_name: Name of metric
        :param metric_values: Metric values to add
        """
        self._results[loop_name][metric_name][i_repeat] = metric_values.flatten()

    def extract_metric_as_array(self, loop_name: str, metric_name: str) -> np.ndarray:
        """
        Returns results over all repeats and iterations for a specific metric and loop name pair

        :param loop_name: Name of loop to return results for
        :param metric_name: Name of metric to extract
        :return: 2-d numpy array of shape (n_repeats x n_iterations)
        """
        return np.array(self._results[loop_name][metric_name])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .warped_bq_model import WarpedBayesianQuadratureModel  # noqa: F401
from .vanilla_bq import VanillaBayesianQuadrature  # noqa: F401



"""
These are emukit model wrappers that contain the specific optimization procedures we found worked well for each model.

The constructor for each model takes X and Y as lists, with each entry of the list corresponding to data for a fidelity
"""
import logging

import GPy
import numpy as np

from ...core.interfaces import IModel
from ...model_wrappers import GPyMultiOutputWrapper
from ...multi_fidelity.convert_lists_to_array import convert_xy_lists_to_arrays
from ...multi_fidelity.kernels import LinearMultiFidelityKernel
from ...multi_fidelity.models import GPyLinearMultiFidelityModel
from ...multi_fidelity.models.non_linear_multi_fidelity_model import (
    NonLinearMultiFidelityModel, make_non_linear_kernels)

_log = logging.getLogger(__name__)


class HighFidelityGp(IModel):
    """
    GP at high fidelity only.

    The optimization is restarted from random initial points 10 times.
    The noise parameter is initialized at 1e-6 for the first optimization round.
    """

    def __init__(self, X, Y):
        kern = GPy.kern.RBF(X[1].shape[1], ARD=True)
        self.model = GPy.models.GPRegression(X[1], Y[1], kernel=kern)
        self.model.Gaussian_noise.variance = 1e-6
        self.name = 'hf_gp'

    def optimize(self):
        _log.info('\n--- Optimization: ---\n'.format(self.name))
        self.model.optimize_restarts(10, robust=True)

    def predict(self, X):
        """
        Predict from high fidelity
        """
        return self.model.predict(X[:, :-1])

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        raise NotImplementedError()

    @property
    def X(self):
        raise NotImplementedError()

    @property
    def Y(self):
        raise NotImplementedError()


class LinearAutoRegressiveModel(IModel):
    """
    Linear model, AR1 in paper. Optimized with noise fixed at 1e-6 until convergence then the noise parameter is freed
    and the model is optimized again
    """
    def __init__(self, X, Y, n_restarts=10):
        """

        :param X: List of training data at each fidelity
        :param Y: List of training targets at each fidelity
        :param n_restarts: Number of restarts during optimization of hyper-parameters
        """
        x_train, y_train = convert_xy_lists_to_arrays(X, Y)
        n_dims = X[0].shape[1]
        kernels = [GPy.kern.RBF(n_dims, ARD=True) for _ in range(len(X))]
        lin_mf_kernel = LinearMultiFidelityKernel(kernels)
        gpy_lin_mf_model = GPyLinearMultiFidelityModel(x_train, y_train, lin_mf_kernel, n_fidelities=len(X))
        gpy_lin_mf_model.mixed_noise.Gaussian_noise.fix(1e-6)
        gpy_lin_mf_model.mixed_noise.Gaussian_noise_1.fix(1e-6)
        if len(Y) == 3:
            gpy_lin_mf_model.mixed_noise.Gaussian_noise_2.fix(1e-6)

        self.model = GPyMultiOutputWrapper(gpy_lin_mf_model, len(X), n_optimization_restarts=n_restarts)
        self.name = 'ar1'
        self.n_fidelities = len(X)

    def predict(self, X):
        """
        Predict from high fidelity
        """
        return self.model.predict(X)

    def optimize(self):
        _log.info('\n--- Optimization: ---\n'.format(self.name))
        self.model.optimize()
        self.model.gpy_model.mixed_noise.Gaussian_noise.unfix()
        self.model.gpy_model.mixed_noise.Gaussian_noise_1.unfix()
        if self.n_fidelities == 3:
            self.model.gpy_model.mixed_noise.Gaussian_noise_2.unfix()
        self.model.optimize()

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        raise NotImplementedError()

    @property
    def X(self):
        raise NotImplementedError()

    @property
    def Y(self):
        raise NotImplementedError()


class NonLinearAutoRegressiveModel(IModel):
    """
    Non-linear model, NARGP in paper
    """
    def __init__(self, X, Y, n_restarts=10):
        x_train, y_train = convert_xy_lists_to_arrays(X, Y)
        base_kernel = GPy.kern.RBF
        kernels = make_non_linear_kernels(base_kernel, len(X), x_train.shape[1] - 1, ARD=True)
        self.model = NonLinearMultiFidelityModel(x_train, y_train, n_fidelities=len(X), kernels=kernels,
                                                 verbose=True, optimization_restarts=n_restarts)
        self.name = 'nargp'

    def predict(self, X):
        """
        Predict from high fidelity
        """
        return self.model.predict(X)

    def optimize(self):
        _log.info('\n--- Optimization: ---\n'.format(self.name))
        self.model.optimize()

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        raise NotImplementedError()

    @property
    def X(self):
        raise NotImplementedError()

    @property
    def Y(self):
        raise NotImplementedError()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License").
#   You may not use this file except in compliance with the License.
#   A copy of the License is located at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   or in the "license" file accompanying this file. This file is distributed
#   on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
#   express or implied. See the License for the specific language governing
#   permissions and limitations under the License.
# ==============================================================================

from setuptools import setup, find_packages
import sys

from emukit.__version__ import __version__

with open('README.md', 'r') as fh:
    long_description = fh.read()

with open('requirements/requirements.txt', 'r') as req:
    requires = req.read().split("\n")

# enforce >Python3 for all versions of pip/setuptools
assert sys.version_info >= (3,), 'This package requires Python 3.'

setup(
    name="emukit",
    version=__version__,
    description='Toolkit for decision making under uncertainty.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/amzn/emukit',
    packages=find_packages(exclude=['test*']),
    include_package_data=True,
    install_requires=requires,
    extras_require={'benchmarking': ['matplotlib']},
    python_requires='>=3',
    license='Apache License 2.0',
    classifiers=(
        # https://pypi.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'
    )
)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import Tuple, List, Optional
import GPy
from scipy.linalg import lapack

from emukit.quadrature.interfaces.standard_kernels import IRBF
from emukit.quadrature.interfaces import IBaseGaussianProcess
from emukit.quadrature.kernels.quadrature_kernels import QuadratureKernel
from emukit.quadrature.kernels.quadrature_rbf import QuadratureRBFIsoGaussMeasure, QuadratureRBFLebesgueMeasure, QuadratureRBFUniformMeasure
from emukit.quadrature.kernels.integration_measures import IntegrationMeasure, IsotropicGaussianMeasure, UniformMeasure


class BaseGaussianProcessGPy(IBaseGaussianProcess):
    """
    Wrapper for GPy GPRegression
    An instance of this can be passed as 'base_gp' to a WarpedBayesianQuadratureModel object.

    Note that the GPy cannot take None as initial values for X and Y. Thus we initialize it with some values. These will
    be re-set in the WarpedBayesianQuadratureModel.
    """
    def __init__(self, kern: QuadratureKernel, gpy_model: GPy.models.GPRegression, noise_free: bool=True):
        """
        :param kern: a quadrature kernel
        :param gpy_model: A GPy GP regression model, GPy.models.GPRegression
        :param noise_free: if False, the observation noise variance will be treated as a model parameter,
        if True it is set to 1e-10, defaults to True
        """
        super().__init__(kern=kern)
        if noise_free:
            gpy_model.Gaussian_noise.constrain_fixed(1.e-10)
        self.gpy_model = gpy_model

    @property
    def X(self) -> np.ndarray:
        return self.gpy_model.X

    @property
    def Y(self) -> np.ndarray:
        return self.gpy_model.Y

    @property
    def observation_noise_variance(self) -> np.float:
        """
        Gaussian observation noise variance
        :return: The noise variance from some external GP model
        """
        return self.gpy_model.Gaussian_noise[0]

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Sets training data in model
        :param X: New training features
        :param Y: New training outputs
        """
        self.gpy_model.set_XY(X, Y)

    def predict(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predictive mean and covariance at the locations X_pred

        :param X_pred: points at which to predict, with shape (number of points, dimension)
        :return: Predictive mean, predictive variances shapes (num_points, 1) and (num_points, 1)
        """
        return self.gpy_model.predict(X_pred, full_cov=False)

    def predict_with_full_covariance(self, X_pred: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predictive mean and covariance at the locations X_pred

        :param X_pred: points at which to predict, with shape (num_points, input_dim)
        :return: Predictive mean, predictive full covariance shapes (num_points, 1) and (num_points, num_points)
        """
        return self.gpy_model.predict(X_pred, full_cov=True)

    def solve_linear(self, z: np.ndarray) -> np.ndarray:
        """
        Solve the linear system G(X, X)x=z for x.
        G(X, X) is the Gram matrix :math:`G(X, X) = K(X, X) + \sigma^2 I`, of shape (num_dat, num_dat) and z is a
        matrix of shape (num_dat, num_obs).

        :param z: a matrix of shape (num_dat, num_obs)
        :return: the solution to the linear system G(X, X)x = z, shape (num_dat, num_obs)
        """
        lower_chol = self.gpy_model.posterior.woodbury_chol
        return lapack.dtrtrs(lower_chol.T, (lapack.dtrtrs(lower_chol, z, lower=1)[0]), lower=0)[0]

    def graminv_residual(self) -> np.ndarray:
        """
        The inverse Gram matrix multiplied with the mean-corrected data

        ..math::

            (K_{XX} + \sigma^2 I)^{-1} (Y - m(X))

        where the data is given by {X, Y} and m is the prior mean and sigma^2 the observation noise

        :return: the inverse Gram matrix multiplied with the mean-corrected data with shape: (number of datapoints, 1)
        """
        return self.gpy_model.posterior.woodbury_vector

    def optimize(self) -> None:
        """ Optimize the hyperparameters of the GP """
        self.gpy_model.optimize()


class RBFGPy(IRBF):
    """
    Wrapper of the GPy RBF kernel

    .. math::
        k(x, x') = \sigma^2 e^{-\frac{1}{2}\frac{\|x-x'\|^2}{\lambda^2}},

    where :math:`\sigma^2` is the `variance' property and :math:`\lambda` is the lengthscale property.
    """

    def __init__(self, gpy_rbf: GPy.kern.RBF):
        """
        :param gpy_rbf: An RBF kernel from GPy with ARD=False
        """
        self.gpy_rbf = gpy_rbf

    @property
    def lengthscale(self) -> np.float:
        return self.gpy_rbf.lengthscale[0]

    @property
    def variance(self) -> np.float:
        return self.gpy_rbf.variance.values[0]

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        The kernel k(x1, x2) evaluated at x1 and x2

        :param x1: first argument of the kernel
        :param x2: second argument of the kernel
        :returns: kernel evaluated at x1, x2
        """
        return self.gpy_rbf.K(x1, x2)

    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel wrt x1 evaluated at pair x1, x2.
        We use the scaled squared distance defined as:

        ..math::

            r^2(x_1, x_2) = \sum_{d=1}^D (x_1^d - x_2^d)^2/\lambda^2

        :param x1: first argument of the kernel, shape = (n_points N, input_dim)
        :param x2: second argument of the kernel, shape = (n_points M, input_dim)
        :return: the gradient of the kernel wrt x1 evaluated at (x1, x2), shape (input_dim, N, M)
        """
        K = self.K(x1, x2)
        scaled_vector_diff = (x1.T[:, :, None] - x2.T[:, None, :]) / self.lengthscale**2
        dK_dx1 = - K[None, ...] * scaled_vector_diff
        return dK_dx1

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """
        gradient of the diagonal of the kernel (the variance) v(x):=k(x, x) evaluated at x

        :param x: argument of the kernel, shape (n_points M, input_dim)
        :return: the gradient of the diagonal of the kernel evaluated at x, shape (input_dim, M)
        """
        num_points, input_dim = x.shape
        return np.zeros((input_dim, num_points))


def create_emukit_model_from_gpy_model(gpy_model: GPy.models.GPRegression,
                                       integral_bounds: Optional[List[Tuple[float, float]]],
                                       measure: Optional[IntegrationMeasure], integral_name: str = '') \
        -> BaseGaussianProcessGPy:
    """
    Wraps a GPy model and returns an emukit quadrature model.

    :param gpy_model: A GPy Gaussian process regression model, GPy.models.GPRegression
    :param integral_bounds: List of D tuples, where D is the dimensionality of the integral and the tuples contain the
    lower and upper bounds of the integral i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]. None means infinite
    bounds.
    :param measure: an integration measure. None means the standard Lebesgue measure is used.
    :param integral_name: the (variable) name(s) of the integral

    :return: emukit model for quadrature with GPy backend (IBaseGaussianProcessGPy)
    """

    # wrap standard kernel
    if isinstance(gpy_model.kern, GPy.kern.RBF):
        if gpy_model.kern.ARD:
            raise ValueError('ARD in GPy rbf kernel must be set to False. {} given.'.format(gpy_model.kern.ARD))
        standard_kernel_emukit = RBFGPy(gpy_model.kern)
    else:
        raise ValueError('Only RBF kernel is supported. Got ', gpy_model.kern.name, ' instead.')

    # check if measure and bounds fit together and wrap combination in a quadrature kernel
    # infinite bounds and Lebesgue measure. Can't do.
    if (integral_bounds is None) and (measure is None):
        raise ValueError('integral_bounds are infinite and measure is standard Lebesgue. Choose either finite bounds '
                         'or an appropriate integration measure.')

    # infinite bounds: Gauss and uniform measure only
    elif (integral_bounds is None) and (measure is not None):
        if isinstance(measure, UniformMeasure):
            quadrature_kernel_emukit = QuadratureRBFUniformMeasure(rbf_kernel=standard_kernel_emukit,
                                                                   integral_bounds=integral_bounds,
                                                                   measure=measure,
                                                                   variable_names=integral_name)
        elif isinstance(measure, IsotropicGaussianMeasure):
            quadrature_kernel_emukit = QuadratureRBFIsoGaussMeasure(rbf_kernel=standard_kernel_emukit,
                                                                    measure=measure,
                                                                    variable_names=integral_name)
        else:
            raise ValueError('Currently only IsotropicGaussianMeasure and UniformMeasure supported with infinite '
                             'integral bounds.')

    # finite bounds, standard Lebesgue measure
    elif (integral_bounds is not None) and (measure is None):
        quadrature_kernel_emukit = QuadratureRBFLebesgueMeasure(rbf_kernel=standard_kernel_emukit,
                                                                integral_bounds=integral_bounds,
                                                                variable_names=integral_name)

    # finite bounds: measure: uniform measure only
    else:
        if isinstance(measure, UniformMeasure):
            quadrature_kernel_emukit = QuadratureRBFUniformMeasure(rbf_kernel=standard_kernel_emukit,
                                                                   integral_bounds=integral_bounds,
                                                                   measure=measure,
                                                                   variable_names=integral_name)
        else:
            raise ValueError('Currently only standard Lebesgue measure (measure=None) is supported with finite '
                             'integral bounds.')

    # wrap the base-gp model
    return BaseGaussianProcessGPy(kern=quadrature_kernel_emukit, gpy_model=gpy_model)



import numpy as np

from emukit.bayesian_optimization.loops.cost_sensitive_bayesian_optimization_loop import \
    CostSensitiveBayesianOptimizationLoop
from emukit.core import ParameterSpace, ContinuousParameter
from emukit.core.acquisition import IntegratedHyperParameterAcquisition
from emukit.core.acquisition import acquisition_per_expected_cost
from emukit.core.loop import FixedIntervalUpdater, SequentialPointCalculator
from emukit.core.loop.loop_state import create_loop_state
from emukit.core.optimization import RandomSearchAcquisitionOptimizer
from emukit.examples.fabolas.continuous_fidelity_entropy_search import ContinuousFidelityEntropySearch
from emukit.examples.fabolas.fabolas_model import FabolasModel


class FabolasLoop(CostSensitiveBayesianOptimizationLoop):

    def __init__(self, space: ParameterSpace,
                 X_init: np.ndarray, Y_init: np.ndarray, cost_init: np.ndarray,
                 s_min: float, s_max: float,
                 update_interval: int = 1,
                 num_eval_points: int = 2000,
                 marginalize_hypers: bool = True):
        """
        Implements FAst Bayesian Optimization for LArge DataSets as described in:

        Fast Bayesian hyperparameter optimization on large datasets
        A. Klein and S. Falkner and S. Bartels and P. Hennig and F. Hutter
        Electronic Journal of Statistics (2017)

        :param space: input space where the optimization is carried out.
        :param X_init: initial data points
        :param Y_init: initial function values
        :param cost_init: initial costs
        :param s_min: smallest possible dataset size
        :param s_max: highest possible dataset size
        :param update_interval:  number of iterations between optimization of model hyper-parameters. Defaults to 1.
        :param num_eval_points: number of points to evaluate the acquisition function
        :param marginalize_hypers: if true, marginalize over the GP hyperparameters
        """

        l = space.parameters
        l.extend([ContinuousParameter("s", s_min, s_max)])  
        extended_space = ParameterSpace(l)

        model_objective = FabolasModel(X_init=X_init, Y_init=Y_init, s_min=s_min, s_max=s_max)
        model_cost = FabolasModel(X_init=X_init, Y_init=cost_init[:, None], s_min=s_min, s_max=s_max)

        if marginalize_hypers:
            acquisition_generator = lambda model: ContinuousFidelityEntropySearch(model_objective, space=extended_space,
                                                                                  target_fidelity_index=len(
                                                                                      extended_space.parameters) - 1)
            entropy_search = IntegratedHyperParameterAcquisition(model_objective, acquisition_generator)
        else:
            entropy_search = ContinuousFidelityEntropySearch(model_objective, space=extended_space,
                                                             target_fidelity_index=len(extended_space.parameters) - 1)

        acquisition = acquisition_per_expected_cost(entropy_search, model_cost)

        model_updater_objective = FixedIntervalUpdater(model_objective, update_interval)
        model_updater_cost = FixedIntervalUpdater(model_cost, update_interval, lambda state: state.cost)

        acquisition_optimizer = RandomSearchAcquisitionOptimizer(extended_space, num_eval_points=num_eval_points)
        candidate_point_calculator = SequentialPointCalculator(acquisition, acquisition_optimizer)

        loop_state = create_loop_state(model_objective.X, model_objective.Y, model_cost.Y)

        super(CostSensitiveBayesianOptimizationLoop, self).__init__(candidate_point_calculator,
                                                                    [model_updater_objective, model_updater_cost],
                                                                    loop_state)



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np

from ...core.parameter_space import ParameterSpace
from ...core.loop.loop_state import LoopState
from ...core.loop.candidate_point_calculators import CandidatePointCalculator
from ..methods.warped_bq_model import WarpedBayesianQuadratureModel
from ...core.optimization.context_manager import ContextManager


class BayesianMonteCarloPointCalculator(CandidatePointCalculator):
    """
    This point calculator implements Bayesian Monte Carlo with simple random sampling.

    The point calculator samples from the probability distribution defined by the integration measure.
    If the integration measure is the standard Lebesgue measure, then this point calculator samples uniformly in a
    box defined by the model (reasonable box).

    C.E. Rasmussen and Z. Ghahramani
    `Bayesian Monte Carlo' Advances in Neural Information Processing Systems 15 (NeurIPS) 2003

    Implemented as described in Section 2.1 of the paper.

    Note that the point calculator does not depend at all on past observations. Thus it is equivalent to sampling
    all points and then fit the model to them. The purpose of the point calculator is convenience, as it can be
    used with the same interface as the active and adaptive learning schemes that depend explicitly or implicitly
    (through hyperparameters) on the previous evaluations.
    """
    def __init__(self, model: WarpedBayesianQuadratureModel, parameter_space: ParameterSpace):
        """
        :param model: A warped Bayesian quadrature model
        """
        self.parameter_space = parameter_space
        self.model = model

        # if measure is probability measure, check if it has sampling capabilities
        if self.model.base_gp.kern.measure is not None:
            if not self.model.base_gp.kern.measure.can_sample:
                raise ValueError("The given probability measure does not support sampling, but Bayesian Monte Carlo "
                                 "requires sampling capability.")

    def compute_next_points(self, loop_state: LoopState, context: dict=None) -> np.ndarray:
        """
        :param loop_state: Object that contains current state of the loop
        :param context: Contains variables to fix and the values to fix them to. The dictionary key is the parameter
        name and the value is the value to fix the parameter to.
        :return: (1 x n_dims) array of next inputs to evaluate the function at
        """

        # Lebesgue measure
        if self.model.base_gp.kern.measure is None:
            if context is None:
                return self.parameter_space.sample_uniform(1)
            else:
                context_manager = ContextManager(self.parameter_space, context)
                samples = context_manager.contextfree_space.sample_uniform(1)
                return context_manager.expand_vector(samples)

        # probability measure
        else:
            if context is None:
                return self.model.base_gp.kern.measure.get_samples(1)
            else:
                context_manager = ContextManager(self.parameter_space, context)
                samples = self.model.base_gp.kern.measure.get_samples(1, context_manager)
                return samples



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import List, Union


class Encoding(object):
    """
    Generic class that represents encodings of categorical variables
    """

    def __init__(self, categories: List[str], encodings: List):
        """
        Initializes an instance of Encoding class

        :param categories: List of categories to encode
        :param encodings: List of corresponding encodings
        """
        self.categories = categories
        self.encodings = np.array(encodings, dtype=float)

        # # As encoding is a static thing and we never add or remove categories
        # # it is fine to maintain multiple copies
        # # We need tuples cause lists are not hashable cannot be used as keys
        # self._encodings_to_categories = {tuple(e): c for e, c in zip(self.encodings, self.categories)}
        # self._categories_to_encodings = {c: e for e, c in zip(self.encodings, self.categories)}

    @property
    def dimension(self) -> int:
        """
        Dimension of the encoding
        """
        return self.encodings.shape[1]

    def round(self, x: np.ndarray) -> np.ndarray:
        """
        Rounds each row in 2d array x to represent one of encodings.

        :param x: A 2d array to be rounded
        :returns: An array where each row represents an encoding
                  that is closest to the corresponding row in x
        """
        if x.ndim != 2:
            raise ValueError("Expected 2d array, got " + str(x.ndim))

        if x.shape[1] != self.dimension:
            raise ValueError("Encoding dimension mismatch, expected {} got {}".format(self.dimension, x.shape[1]))

        x_rounded = []
        for row in x:
            row_rounded = self.round_row(row)
            x_rounded.append(row_rounded)

        return np.row_stack(x_rounded)

    def round_row(self, x_row):
        """
        Rounds the given row. See "round" method docsting for details.

        When subclassing Encoding, it is best to override this method instead of "round".

        :param x_row: A row to round.
        :returns: A rounded row.
        """

        idx = (np.linalg.norm(self.encodings - x_row, axis=1)).argmin()
        return self.encodings[idx].copy()

    def get_category(self, encoding: Union[List, np.ndarray]) -> str:
        """
        Gets the category corresponding to the encoding.

        :param encoding: An encoded value.
        :returns: A category.
        """

        indices = np.where(np.all(self.encodings == np.array(encoding), axis=1))
        if len(indices) == 0 or indices[0].size == 0:
            raise ValueError("Given encoding {} does not correspond to any category".format(encoding))

        category_idx = indices[0][0]
        return self.categories[category_idx]

    def get_encoding(self, category: str) -> List:
        """
        Gets the encoding corresponding to the category.

        :param category: A category.
        :returns: An encoding as a list.
        """
        try:
            index = self.categories.index(category)
        except ValueError:
            raise ValueError("Unknown category {}" + str(category))

        return self.encodings[index].tolist()


class OneHotEncoding(Encoding):
    def __init__(self, categories: List):
        """
        Initializes an instance of OneHotEncoding class
        and generates one hot encodings for given categories.
        Categories are assigned 1's in the order they appear in the provided list.

        An example reference about one hot encoding:
        https://www.kaggle.com/dansbecker/using-categorical-data-with-one-hot-encoding

        :param categories: List of categories to encode
        """
        encodings = []
        for i, _ in enumerate(categories):
            e = [0] * len(categories)
            e[i] = 1
            encodings.append(e)

        super(OneHotEncoding, self).__init__(categories, encodings)

    def round_row(self, x_row):
        """
        Rounds the given row. The highest value is rounded to 1
        all other values are rounded to 0

        :param x_row: A row to round.
        :returns: A rounded row.
        """
        idx = x_row.argmax()
        row_rounded = np.zeros(x_row.shape)
        row_rounded[idx] = 1
        return row_rounded


class OrdinalEncoding(Encoding):
    def __init__(self, categories: List):
        """
        Initializes an instance of OrdinalEncoding class
        and generates ordinal encodings for given categories.
        The encoding is a list of integer numbers [1, .. , n]
        where n is the number of categories.
        Categories are assigned codes in the order they appear in the provided list.

        Note that encoding categories with ordinal encoding is effectively the same as
        treating them as a discrete variable.

        :param categories: List of categories to encode
        """
        encodings = [[i + 1] for i, _ in enumerate(categories)]
        super(OrdinalEncoding, self).__init__(categories, encodings)

    def round_row(self, x_row):
        # since we used just one column for this encoding
        # x_row should contain a single number

        if x_row.shape[0] != 1:
            raise ValueError("Expected a single valued array, got array of {}" + str(x_row.shape))

        x_value = x_row[0]
        if x_value < 1:
            x_value = 1
        if x_value > len(self.categories):
            x_value = len(self.categories)

        rounded_value = int(round(x_value))

        return np.array([rounded_value])



try:
    import torch
    import torch.nn as nn
except ImportError:
    raise ImportError('pytorch is not installed. Please install it by running pip install torch torchvision')

try:
    from pybnn.util.layers import AppendLayer
except ImportError:
    raise ImportError('pybnn is not installed. Please install it by running pip install pybnn')


def get_default_architecture(input_dimensionality: int,
                             classification: bool = False,
                             n_hidden: int = 500) -> torch.nn.Module:
    """
    Defines the architecture that is used for Meta-Surrogate benchmarks.
    In the case of emulating a classification benchmark, we pass the mean prediction through a sigmoid
    to make sure that the mean prediction values are in [0, 1].

    :param input_dimensionality: dimensionality of the benchmark
    :param classification: defined whether we emulate a classification benchmark
    :param n_hidden: number of units in the hidden layer
    :return: nn.Module
    """
    class Architecture(nn.Module):
        def __init__(self, n_inputs, n_hidden=500, classification=False):
            super(Architecture, self).__init__()
            self.fc1 = nn.Linear(n_inputs, n_hidden)
            self.fc2 = nn.Linear(n_hidden, n_hidden)
            self.fc3 = nn.Linear(n_hidden, 1)
            self.sigma_layer = AppendLayer(noise=1e-3)
            self.classification = classification

        def forward(self, x):
            x = torch.tanh(self.fc1(x))
            x = torch.tanh(self.fc2(x))
            x = self.fc3(x)
            if self.classification:
                mean = torch.sigmoid(x)
            else:
                mean = x
            return self.sigma_layer(mean)

    return Architecture(n_inputs=input_dimensionality,
                        n_hidden=n_hidden,
                        classification=classification)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np


class IStandardKernel:
    """
    Interface for a standard kernel k(x, x') that in principle can be integrated
    Inherit from this class to construct wrappers for specific kernels e.g., the rbf
    """

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        The kernel k(x1, x2) evaluated at x1 and x2

        :param x1: first argument of the kernel
        :param x2: second argument of the kernel
        :returns: kernel evaluated at x1, x2
        """
        raise NotImplementedError

    # the following methods are gradients of the kernel
    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel wrt x1 evaluated at pair x1, x2

        :param x1: first argument of the kernel, shape = (n_points N, input_dim)
        :param x2: second argument of the kernel, shape = (n_points M, input_dim)
        :return: the gradient of the kernel wrt x1 evaluated at (x1, x2), shape (input_dim, N, M)
        """
        raise NotImplementedError

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """
        gradient of the diagonal of the kernel (the variance) v(x):=k(x, x) evaluated at x

        :param x: argument of the kernel, shape = (n_points M, input_dim)
        :return: the gradient of the diagonal of the kernel evaluated at x, shape (input_dim, M)
        """
        raise NotImplementedError


class IRBF(IStandardKernel):
    """
    Interface for an RBF kernel
    Inherit from this class to wrap your standard rbf kernel.

    .. math::
        k(x, x') = \sigma^2 e^{-\frac{1}{2}\frac{\|x-x'\|^2}{\lambda^2}},

    where :math:`\sigma^2` is the `variance' property and :math:`\lambda` is the lengthscale property.
    """

    @property
    def lengthscale(self) -> np.float:
        raise NotImplementedError

    @property
    def variance(self) -> np.float:
        raise NotImplementedError

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """
        gradient of the diagonal of the kernel v(x):=k(x, x) evaluated at x.

        :param x: argument of the kernel, shape = (n_points M, input_dim)
        :return: the gradient of the diagonal of the kernel evaluated at x, shape (input_dim, M)
        """
        raise np.zeros((x.shape[1], x.shape[0]))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
import GPy
from typing import List, Tuple, Optional

from emukit.quadrature.methods.vanilla_bq import VanillaBayesianQuadrature
from emukit.quadrature.loop import VanillaBayesianQuadratureLoop
from emukit.quadrature.kernels.integration_measures import IntegrationMeasure
from emukit.model_wrappers.gpy_quadrature_wrappers import create_emukit_model_from_gpy_model


def create_vanilla_bq_loop_with_rbf_kernel(X: np.ndarray, Y: np.ndarray,
                                           integral_bounds: Optional[List[Tuple[float, float]]],
                                           measure: Optional[IntegrationMeasure],
                                           rbf_lengthscale: float=1.0, rbf_variance: float=1.0) \
        -> VanillaBayesianQuadratureLoop:
    """

    :param X: initial training point locations, shape (n_points, input_dim)
    :param Y:  initial training point function values, shape (n_points, 1)
    :param integral_bounds: List of input_dim tuples, where input_dim is the dimensionality of the integral
    and the tuples contain the lower and upper bounds of the integral i.e.,
    [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]. None means infinite bounds.
    :param measure: the integration measure. None means the standard Lebesgue measure is used.
    :param rbf_lengthscale: the lengthscale of the rbf kernel, defaults to 1.
    :param rbf_variance: the variance of the rbf kernel, defaults to 1.
    :return: The vanilla BQ loop
    """

    if (integral_bounds is not None) and (len(integral_bounds) != X.shape[1]):
            D_bounds = len(integral_bounds)
            input_dim = X.shape[1]
            raise ValueError(
                "dimension of integral bounds provided ({}) does not match the input dimension of X ({}).".format(
                    D_bounds, input_dim))
    if rbf_lengthscale <= 0:
        raise ValueError("rbf lengthscale must be positive. The current value is {}.".format(rbf_lengthscale))
    if rbf_variance <= 0:
        raise ValueError("rbf variance must be positive. The current value is {}.".format(rbf_variance))

    gpy_model = GPy.models.GPRegression(X=X, Y=Y, kernel=GPy.kern.RBF(input_dim=X.shape[1],
                                                                      lengthscale=rbf_lengthscale,
                                                                      variance=rbf_variance))

    emukit_model = create_emukit_model_from_gpy_model(gpy_model=gpy_model,
                                                      integral_bounds=integral_bounds,
                                                      measure=measure)
    emukit_method = VanillaBayesianQuadrature(base_gp=emukit_model, X=X, Y=Y)
    emukit_loop = VanillaBayesianQuadratureLoop(model=emukit_method)
    return emukit_loop



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


"""
Contains code for non-linear model multi-fidelity model.

It is based on this paper:

Nonlinear information fusion algorithms for data-efficient multi-fidelity modelling.
P. Perdikaris, M. Raissi, A. Damianou, N. D. Lawrence and G. E. Karniadakis (2017)
http://web.mit.edu/parisp/www/assets/20160751.full.pdf
"""
from typing import Tuple, List, Type

import numpy as np
import GPy

from ...core.interfaces import IModel, IDifferentiable
from ..convert_lists_to_array import convert_y_list_to_array, convert_x_list_to_array


def make_non_linear_kernels(base_kernel_class: Type[GPy.kern.Kern],
                            n_fidelities: int, n_input_dims: int, ARD: bool=False) -> List:
    """
    This function takes a base kernel class and constructs the structured multi-fidelity kernels

    At the first level the kernel is simply:
    .. math
        k_{base}(x, x')

    At subsequent levels the kernels are of the form
    .. math
        k_{base}(x, x')k_{base}(y_{i-1}, y{i-1}') + k_{base}(x, x')

    :param base_kernel_class: GPy class definition of the kernel type to construct the kernels at
    :param n_fidelities: Number of fidelities in the model. A kernel will be returned for each fidelity
    :param n_input_dims: The dimensionality of the input.
    :param ARD: If True, uses different lengthscales for different dimensions. Otherwise the same lengthscale is used
                for all dimensions. Default False.
    :return: A list of kernels with one entry for each fidelity starting from lowest to highest fidelity.
    """

    base_dims_list = list(range(n_input_dims))
    kernels = [base_kernel_class(n_input_dims, active_dims=base_dims_list, ARD=ARD, name='kern_fidelity_1')]
    for i in range(1, n_fidelities):
        fidelity_name = 'fidelity' + str(i + 1)
        interaction_kernel = base_kernel_class(n_input_dims, active_dims=base_dims_list, ARD=ARD,
                                               name='scale_kernel_' + fidelity_name)
        scale_kernel = base_kernel_class(1, active_dims=[n_input_dims], name='previous_fidelity_' + fidelity_name)
        bias_kernel = base_kernel_class(n_input_dims, active_dims=base_dims_list,
                                        ARD=ARD, name='bias_kernel_' + fidelity_name)
        kernels.append(interaction_kernel * scale_kernel + bias_kernel)
    return kernels


class NonLinearMultiFidelityModel(IModel, IDifferentiable):
    """
    Non-linear Model for multiple fidelities. This implementation of the model only handles 1-dimensional outputs.

    The theory implies the training points should be nested such that any point in a higher fidelity exists in all lower
    fidelities, in practice the model will work if this constraint is ignored.
    """

    def __init__(self, X_init: np.ndarray, Y_init: np.ndarray, n_fidelities, kernels: List[GPy.kern.Kern],
                 n_samples=100, verbose=False, optimization_restarts=5) -> None:
        """
        By default the noise at intermediate levels will be fixed to 1e-4.

        :param X_init: Initial X values.
        :param Y_init: Initial Y values.
        :param n_fidelities: Number of fidelities in problem.
        :param kernels: List of kernels for each GP model at each fidelity. The first kernel should take input of
                        dimension d_in and each subsequent kernel should take input of dimension (d_in+1) where d_in is
                        the dimensionality of the features.
        :param n_samples: Number of samples to use to do quasi-Monte-Carlo integration at each fidelity. Default 100
        :param verbose: Whether to output messages during optimization. Defaults to False.
        :param optimization_restarts: Number of random restarts
                                      when optimizing the Gaussian processes' hyper-parameters.
        """

        if not isinstance(X_init, np.ndarray):
            raise TypeError('X_init expected to be a numpy array')

        if not isinstance(Y_init, np.ndarray):
            raise TypeError('Y_init expected to be a numpy array')

        self.verbose = verbose
        self.optimization_restarts = optimization_restarts

        self.n_fidelities = n_fidelities

        # Generate random numbers from standardized gaussian for monte-carlo integration
        self.monte_carlo_rand_numbers = np.random.randn(n_samples)[:, np.newaxis]

        # Make lowest fidelity model
        self.models = []

        self._fidelity_idx = -1

        is_lowest_fidelity = X_init[:, self._fidelity_idx] == 0
        self.models.append(
            GPy.models.GPRegression(X_init[is_lowest_fidelity, :-1], Y_init[is_lowest_fidelity, :], kernels[0]))

        # Make models for fidelities but lowest fidelity
        for i in range(1, self.n_fidelities):
            is_ith_fidelity = X_init[:, self._fidelity_idx] == i
            # Append previous fidelity mean to X
            previous_mean, _ = self._predict_deterministic(X_init[is_ith_fidelity, :-1], i)

            augmented_input = np.concatenate([X_init[is_ith_fidelity, :-1], previous_mean], axis=1)
            self.models.append(GPy.models.GPRegression(augmented_input, Y_init[is_ith_fidelity, :], kernels[i]))

        # Fix noise parameters for all models except top fidelity
        for model in self.models[:-1]:
            model.Gaussian_noise.fix(1e-4)

    def set_data(self, X: np.ndarray, Y: np.ndarray) -> None:
        """
        Updates training data in the model.

        :param X: New training features.
        :param Y: New training targets.
        """
        is_lowest_fidelity = 0 == X[:, -1]
        X_low_fidelity = X[is_lowest_fidelity, :-1]
        Y_low_fidelity = Y[is_lowest_fidelity, :]
        self.models[0].set_XY(X_low_fidelity, Y_low_fidelity)
        for i in range(1, self.n_fidelities):
            is_this_fidelity = i == X[:, -1]
            X_this_fidelity = X[is_this_fidelity, :-1]
            Y_this_fidelity = Y[is_this_fidelity, :]
            previous_mean, _ = self._predict_deterministic(X_this_fidelity, i)
            augmented_input = np.concatenate([X_this_fidelity, previous_mean], axis=1)
            self.models[i].set_XY(augmented_input, Y_this_fidelity)

    @property
    def X(self):
        """
        :return: input array of size (n_points x n_inputs_dims) across every fidelity in original input domain meaning
                 it excludes inputs to models that come from the output of the previous level
        """
        x_list = [self.models[0].X]
        for model in self.models[1:]:
            x_list.append(model.X[:, :-1])
        return convert_x_list_to_array(x_list)

    @property
    def Y(self):
        """
        :return: output array of size (n_points x n_outputs) across every fidelity level
        """
        return convert_y_list_to_array([model.Y for model in self.models])

    @property
    def n_samples(self):
        return self.monte_carlo_rand_numbers.shape[0]

    def predict(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predicts mean and variance at fidelity given by the last column of X

        Note that the posterior isn't Gaussian and so this function doesn't tell us everything about our posterior
        distribution.

        :param X: Input locations with fidelity index appended.
        :returns: mean and variance of posterior distribution at X.
        """

        fidelity = X[:, self._fidelity_idx]

        # Do prediction 1 test point at a time
        variance = np.zeros((X.shape[0], 1))
        mean = np.zeros((X.shape[0], 1))

        for i in range(X.shape[0]):
            sample_mean, sample_var = self._predict_samples(X[[i], :-1], fidelity[i])
            # Calculate total variance and mean from samples
            variance[i, :] = np.mean(sample_var) + np.var(sample_mean)
            mean[i, :] = np.mean(sample_mean)

        return mean, variance

    def get_prediction_gradients(self, X: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        Predicts mean and variance and the gradients of the mean and variance with respect to X.

        :param X: input location.
        :returns: (mean, mean gradient, variance, variance gradient) Gradients will be shape (n_points x (d-1)) because
                  we don't return the gradient with respect to the fidelity index.
        """

        fidelity = X[:, self._fidelity_idx]

        # Initialise vectors
        sample_mean = np.zeros((self.n_samples ** (self.n_fidelities - 1), X.shape[0]))
        d_sample_mean_dx = np.zeros((self.n_samples ** (self.n_fidelities - 1), X.shape[0], X.shape[1] - 1))
        d_sample_var_dx = np.zeros((self.n_samples ** (self.n_fidelities - 1), X.shape[0], X.shape[1] - 1))

        # Iteratively obtain predictions and associated gradients for each input point
        for i in range(X.shape[0]):
            mean, dmean_dx, var, dvar_dx = self._predict_samples_with_gradients(X[[i], :-1], fidelity[i])
            # Assign to outputs
            sample_mean[:, [i]] = mean
            d_sample_mean_dx[:, i, :] = dmean_dx
            d_sample_var_dx[:, i, :] = dvar_dx

        # Calculate means + total variance
        total_mean = np.mean(sample_mean, axis=0, keepdims=True).T
        total_mean_grad = np.mean(d_sample_mean_dx, axis=0)

        # Calculate total variance derivative
        tmp = 2 * np.mean(d_sample_mean_dx * sample_mean[:, :, None], axis=0)
        total_variance_grad = np.mean(d_sample_var_dx, axis=0) + tmp - 2 * total_mean * total_mean_grad

        return total_mean_grad, total_variance_grad

    def _predict_samples(self, X: np.ndarray, fidelity: float):
        """
        Draw samples from model at given fidelity. Returns samples of mean and variance at specified fidelity.
        :param X: Input array without output of previous layer appended.
        :param fidelity: zero based fidelity index.
        :returns sample_mean, sample_variance: mean and variance predictions at input points.
        """

        fidelity = int(fidelity)

        # Predict at first fidelity
        sample_mean, sample_variance = self.models[0].predict(X)

        # Predict at all fidelities up until the one we are interested in
        for i in range(1, fidelity + 1):
            # Draw samples from posterior of previous fidelity
            sample_mean, sample_variance, _ = self._propagate_samples_through_level(X, i, sample_mean, sample_variance)
        return sample_mean, sample_variance

    def _predict_samples_with_gradients(self, X: np.ndarray, fidelity: float):
        """
        Draw samples of mean and variance from model at given fidelity and the gradients of these samples wrt X.

        We calculate the gradients by applying the chain rule as the gradients of each Gaussian process is known wrt
        its inputs.

        :param X: Input array without output of previous layer appended.
        :param fidelity: zero based fidelity index.
        :returns mean, mean gradient, variance, variance gradient: mean and variance predictions at input points.
        """

        fidelity = int(fidelity)

        # Predict at first fidelity
        dsample_mean_dx, dsample_var_dx = self.models[0].predictive_gradients(X)
        dsample_mean_dx = dsample_mean_dx[:, :, 0]
        sample_mean, sample_variance = self.models[0].predict(X)

        for i in range(1, fidelity + 1):
            previous_sample_variance = sample_variance.copy()
            # Predict at all fidelities up until the one we are interested in
            sample_mean, sample_variance, x_augmented = \
                self._propagate_samples_through_level(X, i, sample_mean, sample_variance)
            dsample_mean_dx, dsample_var_dx = \
                self._propagate_samples_through_level_gradient(dsample_mean_dx, dsample_var_dx,
                                                               i, previous_sample_variance, x_augmented)
        return sample_mean, dsample_mean_dx, sample_variance, dsample_var_dx

    def _propagate_samples_through_level(self, X, i_level, sample_mean, sample_variance):
        """
        Sample from the posterior of level i - 1 and propagates these samples through level i.

        :param X: Input array without output of previous layer appended.
        :param i_level: level to push through
        :param sample_mean: mean from previous level
        :param sample_variance: variance from previous level
        """
        # Draw samples from posterior of previous fidelity
        samples = self.monte_carlo_rand_numbers * np.sqrt(sample_variance) + sample_mean.T
        samples = samples.flatten()[:, None]
        # Create inputs for each sample
        x_repeat = np.repeat(X, self.n_samples ** i_level, axis=0)
        # Augment input with mean of previous fidelity
        x_augmented = np.concatenate([x_repeat, samples], axis=1)
        # Predict mean and variance and fidelity i
        sample_mean, sample_variance = self.models[i_level].predict(x_augmented)
        return sample_mean, sample_variance, x_augmented

    def _propagate_samples_through_level_gradient(self, dsample_mean_dx, dsample_var_dx, i_fidelity, sample_variance,
                                                  x_augmented):
        """
        Calculates gradients of sample mean and variance with respect to X when propagated through a level

        :param dsample_mean_dx: Gradients of mean prediction of samples from previous level
        :param dsample_var_dx: Gradients of variance prediction of samples from previous level
        :param i_fidelity: level index
        :param sample_variance: The variance prediction of the samples from the previous level
        :param x_augmented: The X input for this level augmented with the outputs
                            from the previous level as the final column
        """
        # Convert variance derivative to std derivative
        clipped_var = np.clip(sample_variance, 1e-10, np.inf)
        dsample_std_dx = dsample_var_dx / (2 * np.sqrt(clipped_var))
        # Calculate gradients of samples wrt x
        # This calculates a (n_samples**(i-1), n_samples, n_dims) matrix
        tmp = self.monte_carlo_rand_numbers[:, np.newaxis, :] * dsample_std_dx[:, np.newaxis, :]
        dsamples_dx = dsample_mean_dx[np.newaxis, :, :] + tmp
        dsamples_dx_reshaped = np.reshape(dsamples_dx, (self.n_samples ** i_fidelity, dsample_std_dx.shape[1]))

        # Get partial derivatives of mean and variance with respect to
        # both X and output of previous fidelity
        dmean_dx, dvar_dx = self.models[i_fidelity].predictive_gradients(x_augmented)
        dmean_dx = dmean_dx[:, :, 0]

        # Combine partial derivatives to get full derivative wrt X
        dsample_mean_dx = dmean_dx[:, :-1] + dmean_dx[:, [-1]] * dsamples_dx_reshaped
        dsample_var_dx = dvar_dx[:, :-1] + dvar_dx[:, [-1]] * dsamples_dx_reshaped
        return dsample_mean_dx, dsample_var_dx

    def optimize(self) -> None:
        """
        Optimize the full model
        """

        # Optimize the first model
        self.models[0].optimize_restarts(self.optimization_restarts, verbose=self.verbose, robust=True)

        # Optimize all models for all fidelities but lowest fidelity
        for i in range(1, self.n_fidelities):
            # Set new X values because previous model has changed
            previous_mean, _ = self.models[i - 1].predict(self.models[i].X)
            augmented_input = np.concatenate([self.models[i].X[:, :-1], previous_mean], axis=1)
            self.models[i].set_X(augmented_input)

            # Optimize parameters
            self.models[i].optimize_restarts(self.optimization_restarts, verbose=self.verbose, robust=True)

    def get_f_minimum(self) -> np.ndarray:
        """
        Get the minimum of the top fidelity model.
        """
        return np.min(self.models[-1].Y)

    def _predict_deterministic(self, X, fidelity):
        """
        This is a helper function when predicting at points that are in the training set. It is more efficient than
        sampling and is useful when constructing the model.
        """
        # Predict at first fidelity
        mean, variance = self.models[0].predict(X)
        for i in range(1, fidelity):
            # Push samples through this fidelity model
            augmented_input = np.concatenate([X, mean], axis=1)
            mean, variance = self.models[i].predict(augmented_input)
        return mean, variance



from typing import Union

import numpy as np

from emukit.bayesian_optimization.acquisitions import ExpectedImprovement
from emukit.bayesian_optimization.acquisitions.entropy_search import EntropySearch
from emukit.bayesian_optimization.interfaces import IEntropySearchModel
from emukit.core.interfaces import IModel
from emukit.core.parameter_space import ParameterSpace
from emukit.samplers import AffineInvariantEnsembleSampler


class ContinuousFidelityEntropySearch(EntropySearch):
    """
    Entropy search acquisition for continuous fidelity problems. Compared to standard entropy search,
    it computes the information gain only for the distribution of the minimum on the highest fidelity.
    """

    def __init__(self, model: Union[IModel, IEntropySearchModel], space: ParameterSpace,
                 target_fidelity_index: int = None, num_samples: int = 100,
                 num_representer_points: int = 50, burn_in_steps: int = 50):
        """
        :param model: Gaussian process model of the objective function that implements IEntropySearchModel
        :param space: Parameter space of the input domain
        :param target_fidelity_index: The index of the parameter which defines the fidelity
        :param num_samples: Integer determining how many samples to draw for each candidate input
        :param num_representer_points: Integer determining how many representer points to sample
        :param burn_in_steps: Integer that defines the number of burn-in steps when sampling the representer points
        """

        # Find fidelity parameter in parameter space
        if target_fidelity_index is None:
            self.target_fidelity_index = len(space.parameters) - 1
        else:
            self.target_fidelity_index = target_fidelity_index
        self.fidelity_parameter = space.parameters[self.target_fidelity_index]
        self.high_fidelity = self.fidelity_parameter.max

        # Sampler of representer points should sample x location at the highest fidelity
        parameters_without_info_source = space.parameters.copy()
        parameters_without_info_source.remove(self.fidelity_parameter)
        space_without_info_source = ParameterSpace(parameters_without_info_source)

        # Create sampler of representer points
        sampler = AffineInvariantEnsembleSampler(space_without_info_source)

        proposal_func = self._get_proposal_function(model, space)

        super().__init__(model, space, sampler, num_samples, num_representer_points, proposal_func, burn_in_steps)

    def _sample_representer_points(self):
        repr_points, repr_points_log = super()._sample_representer_points()

        # Add fidelity index to representer points
        idx = np.ones((repr_points.shape[0])) * self.high_fidelity
        repr_points = np.insert(repr_points, self.target_fidelity_index, idx, axis=1)
        return repr_points, repr_points_log

    def _get_proposal_function(self, model, space):

        # Define proposal function for multi-fidelity
        ei = ExpectedImprovement(model)

        def proposal_func(x):
            x_ = x[None, :]
            # Map to highest fidelity
            idx = np.ones((x_.shape[0], 1)) * self.high_fidelity

            x_ = np.insert(x_, self.target_fidelity_index, idx, axis=1)

            if space.check_points_in_domain(x_):
                val = np.log(np.clip(ei.evaluate(x_)[0], 0., np.PINF))
                if np.any(np.isnan(val)):
                    return np.array([np.NINF])
                else:
                    return val
            else:
                return np.array([np.NINF])

        return proposal_func



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from ...core.loop.loop_state import create_loop_state
from ...core.loop import OuterLoop, FixedIntervalUpdater, ModelUpdater
from ...core.parameter_space import ParameterSpace
from ..methods import WarpedBayesianQuadratureModel
from ..loop.quadrature_point_calculators import BayesianMonteCarloPointCalculator


class BayesianMonteCarlo(OuterLoop):
    """
    This loop implements Bayesian Monte Carlo with simple random sampling.

    Nodes are samples from the probability distribution defined by the integration measure.
    If the integration measure is the standard Lebesgue measure, then the nodes are sampled uniformly in a
    box defined by the model (reasonable box).

    C.E. Rasmussen and Z. Ghahramani
    `Bayesian Monte Carlo' Advances in Neural Information Processing Systems 15 (NeurIPS) 2003

    Note: implemented as described in Section 2.1 of the paper.

    Note that the sampling scheme does not depend at all on past observations. Thus it is equivalent to sampling
    all points and then fit the model to them. The purpose of the loop is convenience, as it can be used with the same
    interface as the active and adaptive learning schemes that depend explicitly or implicitly (through hyperparameters)
    on the previous evaluations.
    """
    def __init__(self, model: WarpedBayesianQuadratureModel, model_updater: ModelUpdater=None):
        """
        :param model: a warped Bayesian quadrature method, e.g., VanillaBayesianQuadrature
        :param model_updater: Defines how and when the quadrature model is updated if new data arrives. Defaults to
        FixedIntervalUpdater.

        Hint: The default model_updater `FixedIntervalUpdater' updates and optimizes the model after each new sample.
        Since the sampling scheme of Bayesian Monte Carlo does not depend on the model, alternatively, the dummy updater
        NoopModelUpdater may be used which does not update the model. This may save compute time. However, the model
        then needs to be updated manually after the loop ran: i) the collected nodes are stored in model.loop_state.
        ii) call model.set_data(loop_state.X, loop_state.Y) iii) call model.optimize().
        """
        if model_updater is None:
            model_updater = FixedIntervalUpdater(model, 1)

        space = ParameterSpace(model.reasonable_box_bounds.convert_to_list_of_continuous_parameters())
        candidate_point_calculator = BayesianMonteCarloPointCalculator(model, space)
        loop_state = create_loop_state(model.X, model.Y)

        super().__init__(candidate_point_calculator, model_updater, loop_state)

        self.model = model



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


import numpy as np
from typing import List, Tuple, Optional

from ...quadrature.interfaces.standard_kernels import IStandardKernel
from .bounds import BoxBounds
from .integration_measures import IntegrationMeasure


class QuadratureKernel:
    """
    Abstract class for covariance function of a Gaussian process than can be integrated

    Note that each specific implementation of this class must go with a specific standard kernel as input which
    inherits from IStandardKernel. This is because we both want the QuadratureKernel to be backend agnostic
    but at the same time QuadratureKernel needs access to specifics of the standard kernel.

    An example of a specific QuadratureKernel and IStandardKernel pair is QuadratureRBF and IRBF.
    """

    def __init__(self, kern: IStandardKernel, integral_bounds: Optional[List[Tuple[float, float]]],
                 measure: Optional[IntegrationMeasure], variable_names: str='') -> None:
        """
        :param kern: standard emukit kernel
        :param integral_bounds: defines the domain of the integral. List of D tuples, where D is the dimensionality
        of the integral and the tuples contain the lower and upper bounds of the integral
        i.e., [(lb_1, ub_1), (lb_2, ub_2), ..., (lb_D, ub_D)]. None if bounds are infinite
        :param measure: the integration measure. Note that the integral in QuadratureKernel are specific to the
        measure. None means the standard Lebesgue measure is used.
        :param variable_names: the (variable) name(s) of the integral
        """

        # we define reasonable box bounds for the integral because e.g., the optimizer of the acquisition function
        # requires finite bounds. The box is defined by the integration measure. See each integration measure for
        # details. Note that this only affects methods that use this box, e.g. the acqusition optimizers. The integrals
        # in this kernel will have infinite bounds still.
        if (integral_bounds is None) and (measure is None):
            raise ValueError('integral_bounds and measure are both None. At least one of them must be given.')
        if integral_bounds is None:
            reasonable_box_bounds = measure.get_box()
            self.integral_bounds = None
        else:
            reasonable_box_bounds = integral_bounds
            self.integral_bounds = BoxBounds(name=variable_names, bounds=integral_bounds)

        self.reasonable_box_bounds = BoxBounds(name=variable_names, bounds=reasonable_box_bounds)
        self.kern = kern
        self.measure = measure
        self.input_dim = self.reasonable_box_bounds.dim

    def K(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        The kernel k(x1, x2) evaluated at x1 and x2

        :param x1: first argument of the kernel
        :param x2: second argument of the kernel
        :returns: kernel evaluated at x1, x2
        """
        return self.kern.K(x1, x2)

    # the following methods are integrals of a quadrature kernel
    def qK(self, x2: np.ndarray) -> np.ndarray:
        """
        Kernel with the first component integrated out aka. kernel mean

        :param x2: remaining argument of the once integrated kernel, shape (n_points N, input_dim)
        :returns: kernel mean at location x2, shape (1, N)
        """
        raise NotImplementedError

    def Kq(self, x1: np.ndarray) -> np.ndarray:
        """
        Kernel with the second component integrated out aka. kernel mean

        :param x1: remaining argument of the once integrated kernel, shape (n_points N, input_dim)
        :returns: kernel mean at location x1, shape (N, 1)
        """
        raise NotImplementedError

    def qKq(self) -> np.float:
        """
        Kernel integrated over both arguments x1 and x2

        :returns: double integrated kernel
        """
        raise NotImplementedError

    # the following methods are gradients of a quadrature kernel
    def dK_dx1(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel wrt x1 evaluated at pair x1, x2

        :param x1: first argument of the kernel, shape = (n_points N, input_dim)
        :param x2: second argument of the kernel, shape = (n_points M, input_dim)
        :return: the gradient of the kernel wrt x1 evaluated at (x1, x2), shape (input_dim, N, M)
        """
        return self.kern.dK_dx1(x1, x2)

    def dK_dx2(self, x1: np.ndarray, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel wrt x2 evaluated at pair x1, x2
        Note that it is the transposed gradient wrt x1 evaluated at (x2, x1), i.e., the arguments are switched.

        :param x1: first argument of the kernel, shape = (n_points N, N, input_dim)
        :param x2: second argument of the kernel, shape = (n_points N, M, input_dim)
        :return: the gradient of the kernel wrt x2 evaluated at (x1, x2), shape (input_dim, N, M)
        """
        return np.transpose(self.dK_dx1(x1=x2, x2=x1), (0, 2, 1))

    def dKdiag_dx(self, x: np.ndarray) -> np.ndarray:
        """
        gradient of the diagonal of the kernel (the variance) v(x):=k(x, x) evaluated at x

        :param x: argument of the kernel, shape = (n_points M, input_dim)
        :return: the gradient of the diagonal of the kernel evaluated at x, shape (input_dim, M)
        """
        return self.kern.dKdiag_dx(x)

    def dqK_dx(self, x2: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in first argument) evaluated at x2

        :param x2: N points at which to evaluate, shape = (n_points N, N, input_dim)
        :return: the gradient with shape (input_dim, N)
        """
        raise NotImplementedError

    def dKq_dx(self, x1: np.ndarray) -> np.ndarray:
        """
        gradient of the kernel mean (integrated in second argument) evaluated at x1
        :param x1: N points at which to evaluate, shape = (n_points N, N, input_dim)
        :return: the gradient with shape (N, input_dim)
        """
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict

import numpy as np

from .. import ParameterSpace

Context = Dict[str, Any]


class ContextManager:
    """
    Handles the context variables in the optimizer
    """
    def __init__(self, space: ParameterSpace, context: Context):
        """
        :param space: Parameter space of the search problem.
        :param context: Dictionary of variables and their context values.
                        These values are fixed while optimization.
        """
        self.space = space
        self.contextfree_space = ParameterSpace(
            [param for param in self.space.parameters if param.name not in context])
        self.context_space = ParameterSpace(
            [param for param in self.space.parameters if param.name in context])

        # Find indices of context and non context variables
        self.context_idxs = []
        self.context_values = []
        for name in context.keys():
            # Find indices of variable in the input domain
            self.context_idxs += self.space.find_parameter_index_in_model(name)

            # Find encoded values of context variable
            param = self.space.get_parameter_by_name(name)
            if hasattr(param, 'encoding'):
                self.context_values.append(param.encoding.get_encoding(context[name]))
            else:
                self.context_values.append(context[name])

        all_idxs = list(range(space.dimensionality))
        self.non_context_idxs = [idx for idx in all_idxs if idx not in self.context_idxs]

    def expand_vector(self, x: np.ndarray) -> np.ndarray:
        """
        Expand context free parameter vector by values of the context.

        :param x: Context free parameter values as 2d-array
        :return: Parameter values with inserted context values
        """
        if len(self.context_space.parameters) == 0:
            return x
        else:
            x = np.atleast_2d(x)
            x_expanded = np.zeros((x.shape[0], self.space.dimensionality))
            x_expanded[:, np.array(self.non_context_idxs).astype(int)] = x
            x_expanded[:, np.array(self.context_idxs).astype(int)] = self.context_values
            return x_expanded



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0
import logging

import numpy as np

from .. import ParameterSpace
from ..acquisition import Acquisition
from .context_manager import ContextManager

_log = logging.getLogger(__name__)


class AnchorPointsGenerator(object):
    """
    Anchor points are the points from which the optimization of the acquisition function is initialized.

    This base class is for generating such points, and the sub-classes will implement different logic of how the
    points should be selected
    """
    def __init__(self, space: ParameterSpace, num_samples: int):
        """

        :param space: Parameter space describing the input domain of the model
        :param num_samples: Number of initial samples to draw uniformly from the input domain. These points are then
                            evaluated according to logic implemented in the subclasses, and the best are picked
        """
        self.space = space
        self.num_samples = num_samples

    def get_anchor_point_scores(self, X: np.ndarray) -> np.ndarray:
        """
        This abstract method should contain the logic to ascribe scores to different points in the input domain.
        Points with higher scores will be chosen over points with lower scores.

        :param X: (n_samples x n_inputs_dims) arrays containing the points at which to evaluate the anchor point scores
        :return: Array containing score for each input point
        """
        raise NotImplementedError('get_anchor_point_scores is not implemented in the parent class.')

    def get(self, num_anchor: int = 5, context_manager: ContextManager = None) -> np.ndarray:
        """
        :param num_anchor: Number of points to return
        :param context_manager: Describes any fixed parameters in the optimization
        :return: A (num_anchor x n_dims) array containing the anchor points
        """
        # We use the context handler to remove duplicates only over the non-context variables
        if context_manager is not None:
            space = context_manager.contextfree_space
        else:
            space = self.space

        # Generate initial design
        X = space.sample_uniform(self.num_samples)

        # Add context variables
        if context_manager:
            X = context_manager.expand_vector(X)
        scores = self.get_anchor_point_scores(X)
        sorted_idxs = np.argsort(scores)[::-1]
        anchor_points = X[sorted_idxs[:min(len(scores), num_anchor)], :]

        return anchor_points


class ObjectiveAnchorPointsGenerator(AnchorPointsGenerator):
    """
    This anchor points generator chooses points where the acquisition function is highest
    """
    def __init__(self, space: ParameterSpace, acquisition: Acquisition, num_samples: int=1000):
        """
        :param space: The parameter space describing the input domain of the non-context variables
        :param acquisition: The acquisition function
        :param num_samples: The number of points at which the anchor point scores are calculated
        """
        self.acquisition = acquisition
        super(ObjectiveAnchorPointsGenerator, self).__init__(space, num_samples)

    def get_anchor_point_scores(self, X: np.ndarray) -> np.ndarray:
        """
        :param X: The samples at which to evaluate the criterion
        :return:
        """
        are_constraints_satisfied = np.all([np.ones(X.shape[0])] + [c.evaluate(X) for c in self.space.constraints], axis=0)
        scores = np.zeros((X.shape[0],))
        scores[~are_constraints_satisfied] = -np.inf
        scores[are_constraints_satisfied] = self.acquisition.evaluate(X[are_constraints_satisfied, :])[:, 0]
        return scores
